import fprime.fpp.utils.fpp_to_json.node_structs as NodeStructs


class ModuleWriter:
    """
    Writer for an fpp module. Has open() and close() functions to write the module.
    """

    def __init__(self, module: NodeStructs.Module):
        self.module: NodeStructs.Module = module

    def open(self):
        return f"module {self.module.name} {{"

    def close(self):
        return "}"


class TopologyWriter:
    """
    Writer for an fpp topology. Has open() and close() functions to write the topology.
    """

    def __init__(self, topology: NodeStructs.Topology):
        self.topology: NodeStructs.Topology = topology

    def open(self):
        return f"topology {self.topology.name} {{"

    def close(self):
        return "}"


class ConstantWriter:
    """
    Writer for an fpp constant. Has write() function to write the constant out.
    """

    def __init__(self, constant: NodeStructs.Constant):
        self.constant: NodeStructs.Constant = constant

    def write(self):
        return f"constant {self.constant.id} = {self.constant.value}"


class InstanceSpecWriter:
    """
    Writer for an fpp instance spec. Has write() function to write the instance spec out.
    """

    def __init__(self, instance_spec: NodeStructs.InstanceSpec):
        self.instance_spec: NodeStructs.InstanceSpec = instance_spec

    def write(self):
        return f"{self.instance_spec.visibility} instance {self.instance_spec.name}"


class CompInstanceWriter:
    """
    Writer for an fpp component instance. Has write() function to write the component
    instance definition out.
    """

    def __init__(self, comp_instance: NodeStructs.ComponentInst):
        self.comp_instance: NodeStructs.ComponentInst = comp_instance

    def _writePhases(self):
        phaseOpen = False
        part = ""

        for phase in self.comp_instance.phases:
            if (
                self.comp_instance.phases[phase] is not None
                and self.comp_instance.phases[phase] != ""
            ):
                if not phaseOpen:
                    part += f"{{"
                    phaseOpen = True

                part += f'\n    phase Fpp.ToCpp.Phases.{phase} """\n'
                part += self.comp_instance.phases[phase]
                part += '"""'

        if phaseOpen:
            part += "}"

        return part

    def write(self):
        part_1 = f"instance {self.comp_instance.name}: {self.comp_instance.component_name} base id {self.comp_instance.base_id}"

        if (
            self.comp_instance.queue_size is not None
            and self.comp_instance.queue_size != "None"
        ):
            part_1 = part_1 + f" \\ \n    queue size {self.comp_instance.queue_size}"

        if (
            self.comp_instance.stack_size is not None
            and self.comp_instance.stack_size != "None"
        ):
            part_1 = part_1 + f" \\ \n    stack size {self.comp_instance.stack_size}"

        if (
            self.comp_instance.priority is not None
            and self.comp_instance.priority != "None"
        ):
            part_1 = part_1 + f" \\ \n    priority {self.comp_instance.priority}"

        if self.comp_instance.cpu is not None and self.comp_instance.cpu != "None":
            part_1 = part_1 + f" \\ \n    cpu {self.comp_instance.cpu}"

        return part_1 + self._writePhases()


class PortWriter:
    """
    Writer for an fpp port. Has write() function to write the port definition out.
    """

    def __init__(self, port: NodeStructs.Port):
        self.port: NodeStructs.Port = port

    def write(self):
        portParams = ""

        for i in self.port.parameters:
            portParams = portParams + f"{i['name']} : {i['type']},\n"

        return f"port {self.port.name}(\n    {portParams}\n)"


class TopologyImportWriter:
    """
    Writer for an fpp topology import. Has write() function to write the import out.
    """

    def __init__(self, topology_import: NodeStructs.TopologyImport):
        self.topology_import: NodeStructs.TopologyImport = topology_import

    def write(self):
        return f"import {self.topology_import.name}"


class ConnectionGraphWriter:
    """
    Writer for an fpp connection graph. Has write() function to write the connection graph out.
    """

    def __init__(self, connection_graph: NodeStructs.ConnectionGraph):
        self.connection_graph: NodeStructs.ConnectionGraph = connection_graph

    def write(self):
        part = f"connections {self.connection_graph.name} {{"

        for connection in self.connection_graph.connections:
            if (
                connection["source"]["num"] is None
                or connection["source"]["num"] == "None"
            ):
                connection["source"]["num"] = ""

            if connection["dest"]["num"] is None or connection["dest"]["num"] == "None":
                connection["dest"]["num"] = ""

            part = (
                part
                + f"\n    {connection['source']['name']}{connection['source']['num']} -> {connection['dest']['name']}{connection['dest']['num']}"
            )

        return part + "\n}"



import os
import sys
import unittest

from fprime.common.models.serialize.time_type import TimeType

filename = os.path.dirname(__file__)
fprime_path = os.path.join(filename, "../../../../../src")
sys.path.insert(0, fprime_path)


class TimeTypeTestCases(unittest.TestCase):
    def setUp(self):
        self.t0 = TimeType(0, 0, 0, 0)
        self.t1 = TimeType(0, 0, 1, 0)
        self.t15 = TimeType(0, 0, 1, 500000)

    def test_LessThan(self):
        assert self.t0 >= self.t0, "0 is not less than 0"
        assert self.t0 >= 0, "0 is not less than 0"
        assert self.t0 < self.t1, "0 is less than 1"
        assert self.t0 < 1, "0 is less than 1"
        assert self.t0 < self.t15, "0 is less than 1.5"
        assert self.t0 < 1.5, "0 is less than 1.5"

    def test_LessThanEqualTo(self):
        assert self.t0 <= self.t1, "0 is less than 1"
        assert self.t0 <= 1, "0 is less than 1"
        assert self.t0 <= self.t15, "0 is less than 1.5"
        assert self.t0 <= 1.5, "0 is less than 1.5"
        assert self.t0 <= self.t0, "0 is equal to 0"
        assert self.t0 <= 0, "0 is equal to 0"

    def test_EqualTo(self):
        assert self.t0 == self.t0, "0 is equal to 0"
        assert self.t0 == 0, "0 is equal to 0"
        assert self.t0 != self.t1, "0 is not equal to 1"
        assert self.t0 != 1, "0 is not equal to 1"
        assert self.t0 != self.t15, "0 is not equal to 1.5"
        assert self.t0 != 1.5, "0 is not equal to 1.5"
        assert self.t15 == 1.5, "1.5 is not equal to 1.5"

    def test_NotEqualTo(self):
        assert self.t0 == self.t0, "0 is not equal to 0"
        assert self.t0 == 0, "0 is not equal to 0"
        assert self.t0 != self.t1, "0 is equal to 1"
        assert self.t0 != 1, "0 is equal to 1"
        assert self.t0 != self.t15, "0 is equal to 1.5"
        assert self.t0 != 1.5, "0 is equal to 1.5"

    def test_GreaterThan(self):
        assert self.t15 > self.t0, "1.5 is greater than 0"
        assert self.t15 > 0, "1.5 is greater than 0"
        assert self.t15 > self.t1, "1.5 is greater than 1"
        assert self.t15 > 1, "1.5 is greater than 1"
        assert self.t15 <= self.t15, "1.5 is not greater than 1.5"
        assert self.t15 <= 1.5, "1.5 is not greater than 1.5"

    def test_GreaterThanEqualTo(self):
        assert self.t15 >= self.t0, "1.5 is greater than 0"
        assert self.t15 >= 0, "1.5 is greater than 0"
        assert self.t15 >= self.t1, "1.5 is greater than 1"
        assert self.t15 >= 1, "1.5 is greater than 1"
        assert self.t15 >= self.t15, "1.5 is equal to 1.5"
        assert self.t15 >= 1.5, "1.5 is equal to 1.5"

    def test_Add(self):
        # identity property of addition
        t1_1 = self.t1 + self.t0
        assert t1_1 == self.t1, "1 equals 1"
        assert t1_1 is not self.t1, "the two objects are not the same instance"
        # addition examples
        t5 = self.t1 + self.t1 + self.t1 + 2
        assert t5 == 5, "adding 3 timestamps of 1 and a literal 2 is equal to 5"
        t15_1 = self.t0 + 1.5
        assert t15_1 == self.t15, "adding 1.5 to a timestamps of 0 is equal to 1.5"
        t15_1 = 1.5 + self.t0
        assert t15_1 == self.t15, "adding 1.5 to a timestamps of 0 is equal to 1.5"

    def test_Subtract(self):
        # identity property of subtraction
        t1_1 = self.t1 - self.t0
        assert t1_1 == self.t1, "1 equals 1"
        assert t1_1 is not self.t1, "the two objects are not the same instance"
        # some subtraction examples
        t3 = self.t0 + 3
        t15_1 = t3 - 1.5
        assert self.t15 == t15_1, "subtracting 1.5 from three gives 1.5"
        t27 = t3 - 0.3
        assert t27 == 2.7, "subtracting .3 from 3 equals 2.7"
        t15_1 = 3 - self.t15
        assert self.t15 == t15_1, "subtracting 1.5 from three gives 1.5"

    def test_Multiply(self):
        # identity property of subtraction
        t1_1 = self.t1 * self.t1
        assert t1_1 == self.t1, "1 equals 1"
        assert t1_1 is not self.t1, "the two objects are not the same instance"
        # some multiplication examples
        t5 = self.t1 * 5
        assert t5 == 5, "5 times 1 is 5"
        t6 = self.t15 * 4
        assert t6 == 6, "1.5 times 4 is 6"
        t075 = 0.5 * self.t15
        assert t075 == 0.75, "1.5 times .5 is 0.75"

    def test_TrueDivide(self):
        # identity property of division
        t1_1 = self.t1 / self.t1
        assert t1_1 == self.t1, "1 equals 1"
        assert t1_1 is not self.t1, "the two objects are not the same instance"
        # some division examples
        t6 = self.t0 + 6
        t4 = t6 / 1.5
        assert t4 == 4, "dividing 6 by 1.5 is 4"
        t4 = 6 / self.t15
        assert t4 == 4, "dividing 6 by 1.5 is 4"

    def test_FloorDivide(self):
        # identity property of division
        t1_1 = self.t1 // self.t1
        assert t1_1 == self.t1, "1 equals 1"
        assert t1_1 is not self.t1, "the two objects are not the same instance"
        # some division examples
        t6 = self.t0 + 6
        t3 = t6 // 1.7
        assert t3 == 3, "dividing 6 by 1.7 with floor division is 3"
        t3 = 5.5 // self.t15
        assert t3 == 3, "dividing 5.5 by 1.5 with floor division is 3"


if __name__ == "__main__":
    unittest.main()



"""fprime.util.help_text:

Contains the strings and minor constructs used to specify help text in fprime utility. Any new help strings for new
commands should be written here. Implementers should use HelpText.long(key) and HelpText.short(key) to get appropriate
help messages when building argument parsers. In addition, implementers are expected to use the formatter class
RawDescriptionHelpFormatter as these messages are written expecting that new lines will be maintained.

Help text format: like pydoc, the first line is assumed to be a short description (summary) and following lines will be
interpreted as the explanation. Usually a blank line separates the short and long description.

`short description

long paragraph description
`

@author lestarch
"""

import os
import sys

# Attempt to get pkg_resources from "setuptools"
try:
    import pkg_resources

    VERSION = pkg_resources.get_distribution("fprime-tools").version
except ImportError:
    VERSION = "(unknown version)"


EXECUTABLE = os.path.basename(sys.argv[0])
MNEMONIC_HELP_MAP = {
    "global_help": f"""{EXECUTABLE} ({VERSION}): Utility supporting fprime development patterns.

'{EXECUTABLE}' wraps the fprime build system enabling developers to follow standard patterns when developing fprime
applications. Specifically it translates between the developer's context (working directory and supplied flags) to the
build system targets defined for that context. i.e. a developer can change directory into a component directory and run
commands restricted to that component.

Almost all {EXECUTABLE} commands require a valid build cache to run. Thus users should start by running '{EXECUTABLE}
generate' in their desired project before running command. Once a build cache has been generated other commands can
be run. The '--ut' flag sets up a testing build cache and enables unit test commands to be run. An explanation of how
{EXECUTABLE} determines the build cache is included below. More information on creating build caches can be found with:
'{EXECUTABLE} generate --help'.

Examples:

  -- Setup and Build Ref On Default Platform  --
  cd Ref
  {EXECUTABLE} generate
  {EXECUTABLE} build

  -- Setup and Build Ref Unit Tests --
  cd Ref
  {EXECUTABLE} generate --ut
  {EXECUTABLE} build --ut
  {EXECUTABLE} check  # Runs UTs '--ut' is implied

  -- Setup Ref Build for RaspberryPI Platform --
  cd Ref
  {EXECUTABLE} generate raspberrypi
  {EXECUTABLE} build

{EXECUTABLE} uses the working directory and supplied flags as context for all the commands it runs. This context is used
to determine two build properties: build cache (created by 'fprime-util generate') to be used and build target to
execute. The build cache is chosen by recurring upward from the current working directory looking for the first build
cache matching the specified platform. If '--ut' is specified, only UT caches will be selected. The platform is read
from the optional positional argument 'platform'. If not specified on the command line, the 'default_toolchain' setting
in the project's settings.ini file is read. If not specified there, the default toolchain "native" is used.

If a different CMake root is desired, the recursive search behavior can be altered with the '-r/--root'. When
'-r/--root' is specified, a matching build cache will be selected from within the supplied root directory.
'--build-cache' may be supplied to force '{EXECUTABLE}' to use the supplied build cache regardless of other conditions.

Once the build cache has been selected, the build target is chosen. {EXECUTABLE} will run the supplied command for
a chosen directory.  This directory is can be set using the '-p/--path' argument and defaults to the user's current
working directory. This command and directory are translated into the build target to be executed.
e.g. 'cd Ref/SignalGen; {EXECUTABLE} impl' will run the build target 'Ref_SignalGen_impl'. Most users need not care
about the specific build target being run and should think "command run in chosen directory".

There are two other conditions where the described behavior will differ slightly. If the current working directory or
the path specified '-p/--path' points to a deployment, the command supplied will be run against the deployment and will
be run against all components used by that deployment. e.g. 'cd Ref; {EXECUTABLE} build --ut' will build all UTs for
components used by the Ref deployment. Supplying the '--all' flag executes the command on everything defined within the
build system. e.g 'cd Ref; {EXECUTABLE} build --all' builds all components found anywhere in the build system regardless
if they are used by the deployment or merely defined in the build system.

Contextual Examples:

  -- Build SignalGen Component --
  cd Ref/SignalGen
  {EXECUTABLE} build

  -- Build Command Dispatcher UTs for Ref Deployment --
  cd Svc/CmdDispatcher
  {EXECUTABLE} build --ut -r ../../Ref/

  -- Build Command Dispatcher UTs for Ref Deployment (Alternate) --
  cd Ref
  {EXECUTABLE} build --ut -p ../../Svc/CmdDispatcher

  -- Build Everything --
  {EXECUTABLE} build --all

""",
    "subparsers_description": f"""Commands supported by {EXECUTABLE} ({VERSION})

These are the commands that may be run as part of {EXECUTABLE}. These commands wrap various functions of the
development process. One of these commands should be the first argument to an invocation of {EXECUTABLE}. For more
explanation on an individual command, please run '{EXECUTABLE} <command> --help>'.
""",
    "build": f"""Build fprime components, deployments, and unit tests

'{EXECUTABLE} build' builds in the selected directory ('-p/--path' or current working directory). When the selected
directory contains a component, the component is built. When the current directory contains a deployment, the deployment
is built and installed into the build artifact directory. The build cache used to perform this build is described in
'{EXECUTABLE} --help'. Notably supplying the '--ut' flag switches to building the unit tests for components and all
deployment unit tests for deployments.. Supplying the '--all' flag will switch to building everything defined in the
build system.

'{EXECUTABLE} build' requires a build cache to have been generated
'{EXECUTABLE} build --ut' requires a testing build cache to have been generated

Examples:

  -- Build Ref/SignalGen --
  cd Ref/SignalGen
  {EXECUTABLE} build

  -- Build Ref/SignalGen UTs --
  cd Ref/SignalGen
  {EXECUTABLE} build --ut

""",
    "impl": f"""Generate fprime implementation templates.

'{EXECUTABLE} impl' generates the implementation templates for the specified directory ('-p/--path' or current working
directory). Implementation generation is only available for component directories and should not be used on deployments.
Two files will be created <component>.template.cpp and <component>.template.hpp. These contain the fill-in base code for
the component implementation as defined by the component's model. 

When the '--ut' flag is specified the unit test implementation templates are created instead, under <path>/test/ut. The
user should use the Tester.cpp, Tester.hpp and TestMain.cpp files as their fill-in templates. Other files created can be
safely removed as they will be regenerated at build time.

Example:

  -- Create Ref/SignalGen Implementation Templates --
  cd Ref/SignalGen
  {EXECUTABLE} impl

  -- Create Ref/SignalGen Unit Test Implementation Templates --
  cd Ref/SignalGen
  {EXECUTABLE} impl --ut

""",
    "check": f"""Run fprime unit tests with optional test coverage.

'{EXECUTABLE} check' handles the running of unit tests. It can be used on components to run the component's unit tests,
deployments to run deployment unit tests, and with the '--all' flag to run all unit tests found in the build system.
'{EXECUTABLE} check' implies the '--ut' flag and specifying it is redundant.

'{EXECUTABLE} check' can also be supplied the '--coverage' flag. When the '--coverage' flag is supplied, the unit test
is run and 'gcovr' is run on the output to check coverage. The default flags to 'gcovr' are '--print-summary', '--txt',
'--html-details' along with filters to restrict the coverage to the existing component/deployment. Outputs are put in a
subdirectory of the component/deployment titled "coverage".  Additional arguments can be supplied after the
'--pass-through' flag with several caveats: '--pass-through ...' must be the last arguments on the command line, and
all paths supplied to 'gcovr' must be full paths. Users should purge the build cache when switching between
'--coverage --all' and '--coverage' in a deployment cache.

Examples:

  -- Run Ref/SignalGen Unit Tests --
  cd Ref/SignalGen
  {EXECUTABLE} check

  -- Run Ref/SignalGen Unit Tests With Coverage --
  cd Ref/SignalGen
  {EXECUTABLE} check --coverage

""",
    "generate": f"""Generate build caches for the specified deployment

'{EXECUTABLE} generate' is used to setup a build cache to support other commands run by {EXECUTABLE}. Without additional
arguments a build cache will be created for the project in the specified directory ('-p/--path', or current working
directory). It will use the default settings specified in the project's settings.ini file for the target platform,
fprime libraries, etc. Specifying '-r/--root' generates the project at the supplied directory overriding the
directory specified with '-p/--path' and the current working directory. '--ut' should be specified to generate testing
build caches for running unit tests.

Basic Examples:
  -- Generate Ref Build Cache With Default Settings --
  cd Ref
  {EXECUTABLE} generate

  -- Generate Ref Testing Build Cache With Default Settings --
  cd Ref
  {EXECUTABLE} generate --ut

The target platform for the build cache can be overridden using an optional positional argument. This will setup the
build to target the supplied platform using a toolchain configuration and platform support files of the same name.
{EXECUTABLE} searches for these named files in appropriate folders in following locations: cmake directory in the
project's root folder, cmake directories in projects included fprime libraries, cmake directory in fprime framework.
These locations are search in the above order with toolchains expected in a 'toolchain' sub folder and platforms in a
'platform' subdirectory.

Toolchain Example:

  -- Generate Ref Build Cache for Raspberrypi --
  cd Ref
  {EXECUTABLE} generate raspberrypi

Other build properties and variables can be specified using flags of the form '-DVARIABLE_NAME=VALUE'. This allows
developers to switch build options for a specific build generated build. This is often combined with the '--build-cache'
flag to put the specifically configured build cache at a specifically named location. All commands (build, check, etc.)
run against that build cache must also specify the same location with the '--build-cache' flag. '-DVARIABLE_NAME=VALUE'
sets a CMake cache variable of name VARIABLE_NAME to the value VALUE. Most of these variables are described in:
fprime/cmake/options.cmake. Warnings from CMake saying VARIABLE_NAME was set but unused likely means the VARIABLE_NAME
was misspelled.

CMake Flag Examples:

  -- Generate Ref Unit Tests Disabling Autocoder UTs --
  cd Ref
  {EXECUTABLE} generate --ut -DFPRIME_ENABLE_AUTOCODER_UTS=OFF

  -- Generate Ref With Baremetal Scheduler Into Specific Cache --
  cd Ref
  {EXECUTABLE} generate --build-cache `pwd`/build-ref-with-baremetal -DFPRIME_USE_BAREMETAL_SCHEDULER=ON

""",
    "purge": f"""Remove build caches for specified project
    
'{EXECUTABLE} purge' removes build caches for the specified project. It also removes the build_artifacts directory
in that project as well. Caches are searched in pairs: normal build cache, paired unit testing build cache. The
user will be asked to confirm when a build cache is being removed unless the '--force' flag is specified. The platform
will use the settings specified in 'settings.ini' or as specified with the optional positional argument. The
'--build-cache' flag can be used to remove an exact build cache without looking at other build caches.

'{EXECUTABLE} purge' will not remove build caches that do not look like valid build caches unless the '--force' flag is
supplied. In this case, the build cache will be removed with reckless abandon.
    """,
    "info": f"""Print contextual target and build cache information before exiting

'{EXECUTABLE} info' is used to print contextual information to the user before exiting. It will print the available
commands within the current context (working directory, '-p/--path', '-r/--root', etc.) and then exit. Users may
use the info command as a way to test and understand how {EXECUTABLE} is mapping to the context and targets used. info
may also be used to locate the artifact output folders within the build cache in order to see generated files, compiler
outputs, etc.

'{EXECUTABLE} info' will print information for both normal and unit testing builds when possible. If '--build-cache' is
specified then only the information for that build cache will be printed.
""",
    "version-check": f"""Print out toolchain versions to help debugging

'{EXECUTABLE} version-check' is used to display information about toolchain versions. It will output details such as
the installed Python version, the installed Pip version, and version information for all the necessary tools for fprime
before exiting. Users can utilize the version-check command as a tool for debugging and comprehending the dependencies for {EXECUTABLE}.

'{EXECUTABLE} version-check' will print information about toolchain versions for debugging purposes.
""",
    "hash-to-file": f"""Convert FW_ASSERT file id hash to file path

When a project compiles fprime with 'FW_ASSERT_LEVEL' set to 'FW_FILEID_ASSERT' a hash will be emitted in place of the
file path in order to keep assert messages succinct. '{EXECUTABLE} hash-to-file <hash>' will convert this hash value to
the file path of the file that produced the assert. This must be run against the same build cache that created the
binary that tripped the assert. When specifying a non-default platform, the platform comes before the hash value. This
command cannot detect changes to the underlying file data as hashes are computed only from the file name.

Examples:

  -- Map Ref Hash to File --
  cd Ref
  {EXECUTABLE} hash-to-file 0xABCD1234

  -- Map Ref Hash to File for RaspberryPI Platform --
  cd Ref
  {EXECUTABLE} hash-to-file raspberrypi 0xABCD1234
""",
    "new": f"""Generate a new fprime object

'{EXECUTABLE} new' runs a wizard to create new objects in fprime (component, deployment, project).

Usage:
  -- New Deployment --
  Generate a new F' deployment within a F' project. The new deployment command is expected to be ran at the root of
  the project and will create a new deployment in the current directory. Using --overwrite will overwrite only the
  files that are generated by the new deployment.
  -- New Component --
  Generate a new F' component within a F' project. This command prompts for what type of component and what it should
  include. At the end of the generation, the user can chose to automatically add the component to the build system and
  run the implementation generator.
  -- New Subtopology --
  Generate a new F' subtopology. This command prompts for the name of the subtopology, and outputs a folder containing the structure for one. The user can then add the subtopology to their own project depending on where the subtopology is generated.
""",
    "format": f"""Format C/C++ files using clang-format

'{EXECUTABLE} format' uses 'clang-format' to format C/C++ files. It uses the style specified in the .clang-format file
found at the root of the F' framework used by the project (i.e. the 'framework_path' specified in settings.ini).
Files are specified through stdin or by using the '--files [<path/to/file>]*' flag. When reading from stdin, file paths
should be separated by whitespace characters.
Because clang-format will try to format any text file it is fed, '{EXECUTABLE} format' restricts by default the files
that are processed to commonly used C/C++ file extensions (cpp, c++, cxx, cc, c, and their 'h' equivalents). Backup
copies of the formatted files are also created by default.

Note: '{EXECUTABLE} format' requires that the 'clang-format' utility is installed and in the PATH.
More information at https://clang.llvm.org/docs/ClangFormat.html

Examples:

  -- Manual usage --
  {EXECUTABLE} format -f Main.cpp Main.hpp
  {EXECUTABLE} format -f Imu/*
  {EXECUTABLE} format -f *.hpp --pass-through --dry-run

  -- Format all files in a given directory -- 
  {EXECUTABLE} format --dirs Svc Fw Drv
  
  -- From stdin using Git | format all changed files --
  git diff --name-only --relative | {EXECUTABLE} format --stdin

  -- Format all C/C++ files within a module --
  cd Ref/SignalGen
  find . | {EXECUTABLE} format --stdin

  -- With file list --
  {EXECUTABLE} format --stdin < files-to-format.txt
  {EXECUTABLE} format --files OtherFile.cpp --stdin < files-to-format.txt

  -- Allow additional file extension -- 
  {EXECUTABLE} format -f Main.cpp main.py --allow-extension .py


""",
}


class HelpText:
    """
    There are two styles of help text: short (for argument descriptions) and long for full descriptions. This function
    will take a context key (command name, mnemonic, or other string) to lookup the full help text. The short help text
    is the first line of the full help text. This class provides functions for both.
    """

    @staticmethod
    def short(context_key: str, default: str = ""):
        """Short help text for context

        Returns the short help message (first line of the long help).

        Args:
            context_key: mnemonic or other key to look-up appropriate help text
            default: default string to use as the full help text
        Returns:
            short help description
        """
        lines = MNEMONIC_HELP_MAP.get(context_key, default).splitlines()
        return lines[0] if lines else default

    @staticmethod
    def long(context_key: str, default: str = "", version="(unknown version)"):
        """long help text for context

        Returns the long help message.

        Args:
            context_key: mnemonic or other key to look-up appropriate help text
            default: default string to use as the full help text
        Returns:
            long help description
        """
        return MNEMONIC_HELP_MAP.get(context_key, default)



"""
(test) fprime.util.file_util:

Tests the file utility functions for the fprime project.

@author SiboVG
"""

import os
from pathlib import Path
import pytest
from unittest.mock import MagicMock, patch

from fprime.fbuild.builder import Build
from fprime.util.file_util import get_directory_path_relative_to_root


@pytest.fixture
def mock_build():
    """Create a mock Build object for testing"""
    mock = MagicMock(spec=Build)
    return mock


def test_current_dir_is_root(mock_build):
    """Test when current directory is the project root"""
    # Setup
    mock_build.get_settings.return_value = str(Path.cwd())

    # Execute
    result = get_directory_path_relative_to_root(mock_build)

    # Verify
    assert result == ""
    mock_build.get_settings.assert_called_once_with("project_root", None)


@patch("fprime.util.file_util.Path.cwd")
def test_current_dir_is_subdirectory(mock_cwd, mock_build):
    """Test when current directory is a subdirectory of the project root"""
    # Setup
    root_path = Path("/path/to/project")
    current_path = root_path / "subdir1" / "subdir2"

    mock_build.get_settings.return_value = str(root_path)
    mock_cwd.return_value = current_path

    # Execute
    result = get_directory_path_relative_to_root(mock_build)

    # Verify
    assert result.replace(os.sep, "/") == "subdir1/subdir2"


@patch("fprime.util.file_util.Path.cwd")
def test_current_dir_outside_project(mock_cwd, mock_build):
    """Test when current directory is outside the project root"""
    # Setup
    root_path = Path("/path/to/project")
    current_path = Path("/different/path")

    mock_build.get_settings.return_value = str(root_path)
    mock_cwd.return_value = current_path

    # Execute
    result = get_directory_path_relative_to_root(mock_build)

    # Verify
    assert result is None


@patch("fprime.util.file_util.Path.cwd")
def test_with_nested_directories(mock_cwd, mock_build):
    """Test with deeply nested directories"""
    # Setup
    root_path = Path("/path/to/project")
    current_path = root_path / "a" / "b" / "c" / "d"

    mock_build.get_settings.return_value = str(root_path)
    mock_cwd.return_value = current_path

    # Execute
    result = get_directory_path_relative_to_root(mock_build)

    # Verify
    assert result.replace(os.sep, "/") == "a/b/c/d"


def test_with_nonexistent_project_root(mock_build):
    """Test behavior when project_root setting doesn't exist"""
    # Setup
    mock_build.get_settings.return_value = None

    # This should return None
    result = get_directory_path_relative_to_root(mock_build)
    assert result is None


@patch("fprime.util.file_util.Path.cwd")
def test_with_relative_path_in_settings(mock_cwd, mock_build):
    """Test with a relative path in the settings"""
    # Setup
    root_path = Path("project")  # Relative path
    abs_root_path = Path("/path/to/project")  # What it resolves to
    current_path = abs_root_path / "subdir"

    # Mock resolve to return the absolute path
    with patch.object(Path, "resolve", return_value=abs_root_path):
        mock_build.get_settings.return_value = str(root_path)
        mock_cwd.return_value = current_path

        # Execute
        result = get_directory_path_relative_to_root(mock_build)

        # Verify
        assert result == "subdir"



"""
Created on Dec 18, 2014

@author: tcanham

"""

from fprime.util.string_util import format_string_template

from . import array_type
from .type_base import BaseType, DictionaryType
from .type_exceptions import (
    IncorrectMembersException,
    MissingMemberException,
    NotInitializedException,
    TypeMismatchException,
    DeserializeException,
)


class SerializableType(DictionaryType):
    """
    Representation of the Serializable type (comparable to the ANY type)

    The serializable type is a container for other instances of
    BaseType, including itself.

    @param param: typename = "SomeTypeName" string
    To preserve member order, the member argument is a list of members and their types:
    @param param: mem_list = [ ("member",<ref to BaseType>, format string, description), ... ]
    OR mem_list = [ ("member",<ref to BaseType>, format string), ... ].
    The member descriptions can be None
    """

    @classmethod
    def construct_type(cls, name, member_list):
        """Construct a new serializable sub-type

        Constructs a new serializable subtype from the supplied member list and name. Member list may optionally exclude
        description keys, which will be filled with None.

        Args:
            name: name of the new sub-type
            member_list: list of member definitions in form list of tuples (name, type, format string, description)
        """
        # Check that we are dealing with a list
        if not isinstance(member_list, list):
            raise TypeMismatchException(list, type(member_list))
        member_list = [list(item) + ([None] * (4 - len(item))) for item in member_list]
        # Check the validity of the member list
        for member_name, member_type, format_string, description in member_list:
            # Check each of these members for correct types
            if not isinstance(member_name, str):
                raise TypeMismatchException(str, type(member_name))
            if not issubclass(member_type, BaseType):
                raise TypeMismatchException(BaseType, member_type)
            if format_string is not None and not isinstance(format_string, str):
                raise TypeMismatchException(str, type(format_string))
            if description is not None and not isinstance(description, str):
                raise TypeMismatchException(str, type(description))
        return DictionaryType.construct_type(cls, name, MEMBER_LIST=member_list)

    @classmethod
    def validate(cls, val):
        """Validate this object including member list and values"""
        # Ensure that the supplied value is a dictionary
        if not isinstance(val, dict):
            raise TypeMismatchException(dict, type(val))
        if len(val) != len(cls.MEMBER_LIST):
            raise IncorrectMembersException([name for name, _, _, _ in cls.MEMBER_LIST])
        # Now validate each field as defined via the value
        for member_name, member_type, _, _ in cls.MEMBER_LIST:
            try:
                member_val = val[member_name]
            except KeyError:
                raise MissingMemberException(member_name)
            member_type.validate(member_val)

    @property
    def val(self) -> dict:
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This returns such an object.

        :return dictionary of member names to python values of member keys
        """
        if self._val is None:
            return None
        return {
            member_name: self._val.get(member_name).val
            for member_name, _, _, _ in self.MEMBER_LIST
        }

    @val.setter
    def val(self, val: dict):
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This takes such an object and sets the
        member val list from it.

        :param val: dictionary containing python types to key names. This
        """
        self.validate(val)
        self._val = {
            member_name: member_type(val.get(member_name))
            for member_name, member_type, _, _, in self.MEMBER_LIST
        }

    @property
    def formatted_val(self) -> dict:
        """
        Format all the members of dict according to the member_format.
        Note 1: All elements will be cast to str
        Note 2: If a member is an array will call array formatted_val
        :return a formatted dict
        """
        result = {}
        for member_name, _, member_format, _ in self.MEMBER_LIST:
            value_object = self._val[member_name]
            if isinstance(value_object, (array_type.ArrayType, SerializableType)):
                result[member_name] = value_object.formatted_val
            else:
                result[member_name] = format_string_template(
                    member_format, value_object.val
                )
        return result

    def serialize(self):
        """Serializes the members of the serializable"""
        if self._val is None:
            raise NotInitializedException(type(self))
        return b"".join(
            [
                self._val.get(member_name).serialize()
                for member_name, _, _, _ in self.MEMBER_LIST
            ]
        )

    def deserialize(self, data, offset):
        """Deserialize the values of each of the members"""
        new_value = {}
        for member_name, member_type, _, _ in self.MEMBER_LIST:
            try:
                new_member = member_type()
                new_member.deserialize(data, offset)
                new_value[member_name] = new_member
                offset += new_member.getSize()
            except Exception as exc:
                raise DeserializeException(
                    f"Field '{member_name}' [{member_type.__name__}] failed to deserialize: {exc}"
                )
        self._val = new_value

    def getSize(self):
        """The size of a struct is the size of all the members"""
        return sum(self._val.get(name).getSize() for name, _, _, _ in self.MEMBER_LIST)

    @classmethod
    def getMaxSize(cls):
        """Return the maximum size in bytes of the array"""
        return sum(member_type.getMaxSize() for _, member_type, _, _ in cls.MEMBER_LIST)

    def to_jsonable(self):
        """
        JSONable type for a serializable
        """
        members = {}
        for member_name, member_value, member_format, member_desc in self.MEMBER_LIST:
            value = (
                {"value": None}
                if self._val is None
                else self._val.get(member_name).to_jsonable()
            )
            members[member_name] = {"format": member_format, "description": member_desc}
            members[member_name].update(value)
        return members

    def items(self):
        """Allow dictionary like iteration"""

        def items_generator():
            """Generator of items"""
            for member_name, _, _, _ in self.MEMBER_LIST:
                yield member_name, self._val.get(member_name)

        return items_generator()



"""fprime.fpp.visualize: Command line targets for fprime-util visualize

@author thomas-bc
"""

import argparse
import os
import shutil
import subprocess
import tempfile
import webbrowser
from pathlib import Path
from typing import Callable, Dict, List, Tuple

from fprime.fpp.common import FppUtility

try:
    from fprime_visual.flask.app import construct_app
except ImportError:
    construct_app = None


def run_fprime_visualize(
    build: "Build",
    parsed: argparse.Namespace,
    _: Dict[str, str],
    __: Dict[str, str],
    ___: List[str],
):
    """Run pipeline of utilities to generate visualization. This includes:
    - fpp-to-layout
    - fpl-layout
    - start fprime-visual Flask app to serve visualization

    Args:
        build: build directory output
        parsed: parsed input arguments
        _: unused cmake_args
        __: unused make_args
        ___: unused pass-through arguments
    """
    if construct_app is None:
        raise ModuleNotFoundError(
            "fprime-visual is not installed. Please install with `pip install fprime-visual`"
        )

    if not shutil.which("fpl-layout"):
        raise FileNotFoundError(
            "fpl-layout is not installed. Please install with `pip install fprime-fpp>1.2.0`"
        )

    # Set up working directory using specified directory, or create a temporary one
    if parsed.working_dir:
        viz_cache_base = Path(parsed.working_dir).resolve()
    else:
        viz_cache_base = Path(
            tempfile.TemporaryDirectory(prefix="fprime-visual-").name
        ).resolve()

    combined_env = os.environ.copy()
    combined_env.update(build.settings.get("environment", {}))

    # Set sub-paths for different types of generated files
    txt_cache = (viz_cache_base / "txt").resolve()
    try:
        txt_cache.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        raise PermissionError(
            f"Unable to write to {viz_cache_base.resolve()}. Use --working-dir to set a different location."
        )

    # Run fpp-to-layout
    FppUtility("fpp-to-layout").execute(
        build,
        parsed.path,
        args=(
            {},
            ["--directory", str(txt_cache)],
        ),
    )
    topology_dirs = list(txt_cache.glob("*Layout"))
    source_dirs = []
    for dir in topology_dirs:
        layout_txt_file_match = list(dir.glob("*.txt"))
        if not layout_txt_file_match:
            raise Exception(f"Did not generate any '*.txt' layout files")
        topology_name = dir.name.replace("Layout", "")
        viz_cache = viz_cache_base / topology_name
        topology_json = viz_cache / f"{topology_name}Topology.json"
        # keep track of all connections in the topology (used in topology layout JSON file)
        topology_connections = ""
        try:
            viz_cache.mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError(
                f"Unable to write to {viz_cache_base.resolve()}. Use --working-dir to set a different location."
            )
        for layout_txt in layout_txt_file_match:
            print(f"Generated layout TXT file: {layout_txt.resolve()}")
            connection_graph_json = viz_cache / f"{layout_txt.stem}.json"
            # Execute: fpl-layout < ConnectionGraph.txt > ConnectionGraph.json
            with open(connection_graph_json.resolve(), "w") as json_file:
                with open(layout_txt.resolve(), "r") as txt_file:
                    txt_contents = txt_file.read()
                    topology_connections += txt_contents
                    subprocess.run(
                        ["fpl-layout"],
                        stdout=json_file,
                        input=txt_contents.encode(),
                        check=True,
                        env=combined_env,
                    )
        # Generate layout JSON for entire topology (all connections graphs in one layout)
        with open(topology_json.resolve(), "w") as json_file:
            subprocess.run(
                ["fpl-layout"],
                stdout=json_file,
                input=topology_connections.encode(),
                check=True,
                env=combined_env,
            )
        source_dirs.append(viz_cache)
    source_resolved = [str(source.resolve()) for source in source_dirs]
    print("[INFO] Starting fprime-visual server...")
    print(f"[INFO] Serving files in {source_resolved}")
    config = {"SOURCE_DIRS": source_resolved}
    app = construct_app(config)
    try:
        webbrowser.open(f"http://localhost:{parsed.gui_port}")
        app.run(port=parsed.gui_port)
    except KeyboardInterrupt:
        print("[INFO] CTRL-C received. Exiting.")
    return 0


def add_fpp_viz_parsers(
    subparsers, common: argparse.ArgumentParser
) -> Tuple[Dict[str, Callable], Dict[str, argparse.ArgumentParser]]:
    """Sets up the fprime-viz command line parsers

    Creates command line parsers for fprime-viz commands and associates these commands to processing functions for those fpp
    commands.

    Args:
        subparsers: subparsers to add to
        common: common parser for all fprime-util commands

    Returns:
        Tuple of dictionary mapping command name to processor, and command to parser
    """
    viz_parser = subparsers.add_parser(
        "visualize",
        help="Visualize FPP model in a web GUI",
        parents=[common],
        add_help=False,
    )
    viz_parser.add_argument(
        "--gui-port",
        help="Set the GUI port for fprime-visual [default: %(default)s]",
        required=False,
        default=7000,
    )
    viz_parser.add_argument(
        "--working-dir",
        help="Set the directory to store layout files in (default to ephemeral location)",
        required=False,
    )
    return {"visualize": run_fprime_visualize}, {"visualize": viz_parser}



"""Test CMake interaction

This file contains a set of tests for validating CMake interaction on the back-end of fprime-util. These tests poke at
the CMakeHandler component.
"""

import json
import os
import shutil
import sys
import tempfile


from contextlib import contextmanager
from pathlib import Path
from unittest.mock import patch, call

import pytest
from fprime.fbuild.cmake import (
    CMakeHandler,
    CMakeExecutionException,
    CMakeInconsistentCacheException,
)


@contextmanager
def temporary_symbolic_link_context(link_name: str):
    """Context manager to create a symbolic link to 'echoer.py' in a temporary directory and add the to sys.path.

    This context manager is designed to facilitate testing of command-line utilities. Specifically, it:
    1. Creates a temporary directory.
    2. Generates a symbolic link inside the temporary directory that points to a file named 'echoer.py' located next to
       this script.
    3. Adds the temporary directory to `sys.path`, ensuring the symbolic link can be imported or executed during testing

    The purpose of this setup is to intercept calls to command-line utility sub commands during testing and echo back
    the context and environment. This allows developers to verify that the correct commands are run by the
    code-under-test, (sub commands within `fprime-util`. By leveraging the `echoer.py` script, testers can capture and
    validate the execution flow without running actual commands, ensuring consistent and isolated test environments.

    On cleanup, the symbolic link, the temporary directory, and its addition to `sys.path` are all reverted.

    Args:
        link_name (str): The name of the symbolic link to create.

    Yields:
        list: list that will contain the results harvested from files **after** the context manager is done

    Raises:
        OSError: If the symbolic link creation fails.
    """
    script_dir = Path(__file__).resolve().parent
    target_path = script_dir / "echoer.py"
    temp_dir = Path(tempfile.mkdtemp())
    link_path = temp_dir / link_name
    results = []
    try:
        # Create the symbolic link
        link_path.symlink_to(target_path)

        # Add the temporary directory to sys.path
        sys.path.insert(0, str(temp_dir))
        os.environ["PATH"] = f'{temp_dir}:{os.environ["PATH"]}'

        # Yield the symbolic link path to the caller
        yield results

        faux_files = list(link_path.parent.glob(f"faux-{link_path.name}-*.json"))
        faux_files.sort()
        for faux_file in faux_files:
            print("Reading Invocation File:", faux_file)
            with faux_file.open("r") as file:
                results.append(json.load(file))

    finally:
        # Remove the symbolic link if it exists
        if link_path.is_symlink():
            link_path.unlink()

        # Remove the temporary directory
        if temp_dir.is_dir():
            shutil.rmtree(temp_dir)

        sys.path.remove(str(temp_dir))
        new_list = os.environ["PATH"].split(":")
        new_list.remove(str(temp_dir))
        os.environ["PATH"] = ":".join(new_list)


def assert_valid_outputs(
    outputs, program="cmake", arguments=None, directory=None, environment=None
):
    """Assert that the outputs of a program invocation are valid.

    This function validates a list of program outputs, ensuring that the program has been invoked correctly
    with expected directories, environment variables, and arguments. The validation occurs as follows:

    1. Checks that there is at least one invocation of the program.
    2. Ensures that required keys ("environment", "cwd", "arguments") exist in each output.
    3. Verifies the working directory (`cwd`) if a `directory` argument is provided.
    4. Confirms the program's environment contains the expected keys and values.
    5. Validates that the program name and arguments match the supplied `arguments` list.

    Args:
        outputs (list): A list of outputs of the echo invocation, where each output is a dictionary containing:
            - "environment" (dict): The environment variables used for the program invocation.
            - "cwd" (str): The working directory where the program was run.
            - "arguments" (list): The arguments passed to the program.
        program (str): The name of the program being validated (default is "cmake").
        arguments (list, optional): Expected list of arguments to the program, excluding the program name.
        directory (str, optional): Expected working directory where the program should be executed.
        environment (dict, optional): A dictionary of environment variables expected in the program's environment.

    Raises:
        AssertionError: If any of the outputs fail validation.
    """
    assert outputs, f"Expected at least one invocation of {program}"

    # Validate output from CLI override
    for key in ["environment", "cwd", "arguments"]:
        for index, output in enumerate(outputs):
            assert (
                key in output
            ), f"{key} not written in the {index} invocation of {program}"

    for output in outputs:
        # If directory is supplied, standardize and check it
        if directory is not None:
            directory = str(Path(directory).resolve())
            assert (
                output["cwd"] == directory
            ), f"{program} not run in the correct directory"

        # If environment is supplied, check for the presence of each key and value
        if environment is not None:
            for key, value in environment.items():
                assert (
                    key in output["environment"]
                ), f"Environment does not contain {key}"
                assert (
                    output["environment"][key] == environment[key]
                ), f"Environment {key} value incorrect"

    # Check program argument and if supplied, the other arguments
    assert Path(output["arguments"][0]).name == program
    if arguments is not None:
        # Strip program off arguments and compare remainder
        output["arguments"] = output["arguments"][1:]
        assert (
            output["arguments"] == arguments
        ), f"Arguments supplied to {program} are incorrect"


@pytest.fixture
def cmake_handler():
    """Fixture to initialize the CMakeHandler"""
    return CMakeHandler()


def test_run_cmake(cmake_handler):
    """Tests the _run_cmake function of the CMakeHandler.

    This function verifies that the `_run_cmake` method of the `CMakeHandler`
    behaves as expected when provided with a basic set of arguments and a custom
    environment. It ensures that the method produces the correct standard output
    and standard error, and it validates the outputs in the specified directory.

    Steps:
        1. A basic set of arguments and environment variables are defined.
        2. A temporary symlink context is created to capture outputs.
        3. The `_run_cmake` function is executed within a temporary directory.
        4. Outputs are validated using `assert_valid_outputs`.
        5. Assertions are made to ensure correct stdout and stderr outputs.

    Raises:
        AssertionError: If the output does not contain the expected information
            or contains unexpected errors.

    Example:
        The function performs the following:
            - Runs `_run_cmake` with arguments ["a", "b", "c"].
            - Ensures no `[ERROR]` appears in the stderr output.
            - Verifies that stdout and stderr contain expected `[INFO]` messages.

    Assertions:
        - Captured arguments, environment, and working directory are as specified
        - `[INFO] Running echoer program (stdout)` appears in the standard output.
        - `[INFO] Running echoer program (stderr)` appears in the standard error.
        - No `[ERROR]` is present in the stderr output.
    """
    arguments = ["a", "b", "c"]
    environment = {"TEST": "TEST_VALUE"}
    with temporary_symbolic_link_context("cmake") as outputs:
        with tempfile.TemporaryDirectory() as test_directory:
            stdout, stderr = cmake_handler._run_cmake(
                arguments=arguments,
                workdir=test_directory,
                environment=environment,
                print_output=False,  # Capture output for test validation
            )
    # Read only mode
    arguments = ["-N"] + arguments
    assert_valid_outputs(
        outputs, directory=test_directory, arguments=arguments, environment=environment
    )
    assert (
        "[INFO] Running echoer program (stdout)\n" in stdout
    ), "Correct standard out not produced"
    assert (
        "[INFO] Running echoer program (stderr)\n" in stderr
    ), "Correct standard error not produced"
    assert "[ERROR]" not in "".join(stderr), "Echoer produced error"


def test_run_cmake_editable(cmake_handler):
    """Tests the writeable _run_cmake function of the CMakeHandler

    This function verifies that the `_run_cmake` method of the `CMakeHandler`
    behaves as expected when provided with a basic set of arguments and a custom
    environment. It ensures that the method produces the correct standard output
    and standard error, and it validates the outputs in the specified directory.

    This step differs by ensuring that _run_cmake does not supply -N when it is a
    writeable operation.

    See test_run_cmake for more information

    Assertions:
        - Captured arguments, environment, and working directory are as specified
        - `[INFO] Running echoer program (stdout)` appears in the standard output.
        - `[INFO] Running echoer program (stderr)` appears in the standard error.
        - No `[ERROR]` is present in the stderr output.
    """
    arguments = ["a", "b", "c"]
    environment = {"TEST": "TEST_VALUE"}
    with temporary_symbolic_link_context("cmake") as outputs:
        with tempfile.TemporaryDirectory() as test_directory:
            stdout, stderr = cmake_handler._run_cmake(
                arguments=arguments,
                workdir=test_directory,
                environment=environment,
                write_override=True,
                print_output=False,  # Capture output for test validation
            )
    assert_valid_outputs(
        outputs, directory=test_directory, arguments=arguments, environment=environment
    )
    assert (
        "[INFO] Running echoer program (stdout)\n" in stdout
    ), "Correct standard out not produced"
    assert (
        "[INFO] Running echoer program (stderr)\n" in stderr
    ), "Correct standard error not produced"
    assert "[ERROR]" not in "".join(stderr), "Echoer produced error"


@patch("fprime.fbuild.cmake.CMakeHandler.get_cmake_module")
@patch("fprime.fbuild.cmake.CMakeHandler.validate_cmake_cache")
@patch("fprime.fbuild.cmake.CMakeHandler._run_cmake")
def mock_execute_known_target(
    cmake_handler,
    setup_callback,
    mock_run_cmake,
    mock_validate_cmake_cache,
    mock_get_cmake_module,
    **kwargs,
):
    """Run the nominal execution of execute_known_target

    Mocks out the _run_cmake, validate_cmake_cache, and get_cmake_module calls and then calls execute_known_target on
    the provided cmake_handler. `setup_callback` is called to configure the behavior for _run_cmake and
    validate_cmake_cache. Any additional arguments to the call are provided by kwargs.

    This will return the mocked_cmake_run call and the expected "call" object provided to it. The expected_target is
    calculated to reflect CMake structure and is adjusted if top_target is in **kwargs.  Environment will be checked if
    supplied in **kwargs.

    Args:
        cmake_handler: cmake handler to use
        setup_callback: callback taking arguments "mock_run_cmake", "mock_validate_cmake_cache" allowing more setup
        ...: provided by mocking utilities

    Return:
        tuple of (mock_run_cmake, expected_call)

    """
    build_dir = "/fake/build/dir"
    target = "my_valid_target"
    module = "CMake_Module"
    expected_target = (
        f"{target}" if kwargs.get("top_target", False) else f"{module}_{target}"
    )

    mock_get_cmake_module.return_value = module
    if setup_callback is not None:
        setup_callback(mock_run_cmake, mock_validate_cmake_cache)

    cmake_handler.execute_known_target(
        target, build_dir, None, print_output=False, **kwargs
    )
    expected_environment = kwargs.get("environment", {})
    expected_make_args = [
        f"{key}={value}" for key, value in kwargs.get("make_args", {}).items()
    ]

    # Verify that _run_cmake was called with the correct arguments
    expected_call = call(
        ["--build", build_dir, "--target", expected_target, "--"] + expected_make_args,
        write_override=True,
        environment=expected_environment,
        print_output=False,
    )
    return mock_run_cmake, expected_call


def test_execute_known_valid_target(cmake_handler):
    """Test execute_known_target with a valid target"""
    mock_run_cmake, expected_call = mock_execute_known_target(cmake_handler, None)
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])


@patch("fprime.fbuild.cmake.CMakeHandler.cmake_refresh_cache")
def test_refresh_and_execute_known_target_valid(
    mock_cmake_refresh_cache, cmake_handler
):
    """Test execute_known_target triggering a cache refresh and reattempt"""

    def extra_setup(mock_run_cmake, _):
        """Perform extra setup to the basic mock calls"""
        mock_run_cmake.side_effect = CMakeExecutionException(
            ["No rule to make target"], ["No rule to make target"], False
        )

    with pytest.raises(CMakeExecutionException):
        mock_run_cmake, expected_call = mock_execute_known_target(
            cmake_handler, extra_setup
        )
        mock_cmake_refresh_cache.assert_called_once()
        mock_run_cmake.assert_has_calls([expected_call, expected_call])


def test_execute_known_global_target_valid(cmake_handler):
    """Test execute_known_target with a valid target at global scope"""
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, top_target=True
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])


def test_execute_known_cache_invalid(cmake_handler):
    """Test execute_known_target triggering an invalid cache"""

    def extra_setup(_, mock_validate_cmake_cache):
        """Perform extra setup to the basic mock calls"""
        mock_validate_cmake_cache.side_effect = CMakeInconsistentCacheException(
            "Module", "A", "B"
        )

    with pytest.raises(
        CMakeInconsistentCacheException,
        match="Module to be set to 'A', was actually set to 'B'",
    ):
        mock_run_cmake, expected_call = mock_execute_known_target(
            cmake_handler, extra_setup
        )
        # Verify that _run_cmake was called with the correct arguments
        mock_run_cmake.assert_not_called()


def test_execute_known_valid_targets_with_environment(cmake_handler):
    """Test execute_known_target on valid local and global targets with supplied environment"""
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, environment={"MY_ENV_VAR": "VALUE1"}
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, environment={"MY_ENV_VAR": "VALUE1"}, top_target=True
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])


def test_execute_known_valid_targets_with_make_args(cmake_handler):
    """Test execute_known_target on valid local and global targets with supplied make_args

    Make args are supplied as command line arguments after a -- when CMake is called.
    """
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, make_args={"--make-arg": "VALUE1"}
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, make_args={"--make-arg": "VALUE1"}, top_target=True
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])


def test_execute_known_valid_targets_with_cmake_args(cmake_handler):
    """Test execute_known_target on valid local and global targets with supplied cmake_args

    CMake args **are not** supplied to the cmake call, but are used only to validate the cache.
    """
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, make_args={"--make-arg": "VALUE1"}
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])
    mock_run_cmake, expected_call = mock_execute_known_target(
        cmake_handler, None, cmake_args={"--cmake-arg": "VALUE1"}, top_target=True
    )
    mock_run_cmake.assert_called_once()
    mock_run_cmake.assert_has_calls([expected_call])


@patch("fprime.fbuild.cmake.CMakeHandler._is_noop_supported")
@patch("fprime.fbuild.cmake.CMakeHandler.execute_known_target")
def cmake_refresh_cache(
    noop_supported,
    cmake_handler,
    mock_execute_known_target,
    mock_is_noop_supported,
    **kwargs,
):
    """Test the CMake cache refresh calls

    Tests that cmake_refresh_cache properly operates in calling execute_known_target under different circumstances. This
    helper assumes that this code will not do a full refresh.
    """
    mock_is_noop_supported.return_value = noop_supported
    assert not kwargs.get(
        "full", False
    ), "CMake refresh cache helper cannot be used with 'full'"

    cmake_handler.cmake_refresh_cache("/some/build/dir", **kwargs)
    mock_execute_known_target.assert_called_once_with(
        "noop" if noop_supported else "refresh_cache",  # Expected target to use
        "/some/build/dir",  # Build directory
        None,  # Path unused for global target
        top_target=True,  # Global target
        full_cache_rebuild=True,  # Prevents infinite recursion, fallback for if the refresh target fails
        print_output=True,  # User will want to know
        environment=kwargs.get("environment", {}),
    )


def test_refresh_cache_noop(cmake_handler):
    """Test the refresh cache handler with older 'noop' target"""
    cmake_refresh_cache(True, cmake_handler)


def test_refresh_cache_refresh(cmake_handler):
    """Test the refresh cache handler with newer 'refresh_cache' target"""
    cmake_refresh_cache(False, cmake_handler)


def test_refresh_cache_noop_with_environment(cmake_handler):
    """Test the refresh cache handler with older 'noop' target"""
    cmake_refresh_cache(True, cmake_handler, environment={"MY_ENV_VAR": "VALUE1"})


@patch("os.makedirs")
@patch("os.path.exists")
@patch("pathlib.Path.touch")
@patch("fprime.fbuild.cmake.CMakeHandler.cmake_validate_source_dir")
@patch("fprime.fbuild.cmake.CMakeHandler._run_cmake")
def generate_build(
    arguments,
    exists,
    cmake_handler,
    mock_run_cmake,
    mock_validate_source_dir,
    mock_touch,
    mock_path_exists,
    mock_makedirs,
    **kwargs,
):
    """Run the generate_build function with mocked dependencies.

    This will run a mocked version of generate_build for the purposes of testing the function. This function supplies
    the standard inputs and asserts the basic calls to the underlying functions (e.g. _run_cmake) through the mocking
    pattern.

    Args:
        arguments: tuple of (dictionary, list of strings) representing the passed-in and output arguments to cmake
        cmake_handler: object under test (from harness)
        ...: mocked out functions
    """
    source_dir = "/fake/source/dir"
    build_dir = "/fake/build/dir"
    kwargs["args"], expected_cmake_args = arguments

    # Mock os.path.exists to return False, simulating that the build_dir does not exist
    mock_path_exists.return_value = exists

    # Call the generate_build function
    cmake_handler.generate_build(source_dir, build_dir, **kwargs)

    # Asset that the touch method was called once
    mock_touch.assert_called_once()

    # Assert that cmake_validate_source_dir was called with the correct source_dir
    mock_validate_source_dir.assert_called_once_with(source_dir)

    # Assert that os.makedirs was called to create the build directory only when it does not exist
    if exists:
        mock_makedirs.assert_not_called()
    else:
        mock_makedirs.assert_called_once_with(build_dir)
    # Assert that _run_cmake was called with the correct arguments
    expected_cmake_args = ["-S", str(Path(source_dir).resolve())] + expected_cmake_args
    mock_run_cmake.assert_called_once_with(
        expected_cmake_args,
        workdir=build_dir,
        print_output=True,
        write_override=True,
        environment=kwargs.get("environment", None),
    )


def test_generate(cmake_handler):
    """Test basic generate build call

    Tests generate_build with no arguments, off-nominal conditions, nor environment. Delegates to the generate_build.
    """
    generate_build(({}, []), False, cmake_handler)


def test_generate_with_args(cmake_handler):
    """Test basic generate build call with supplied cmake arguments

    Tests generate_build with arguments but no off-nominal conditions, nor environment. Delegates to the generate_build.
    """
    generate_build(
        (
            {"CMAKE_BUILD_TYPE": "Debug", "--some-flag": "some-value"},
            ["-DCMAKE_BUILD_TYPE=Debug", "--some-flag=some-value"],
        ),
        False,
        cmake_handler,
    )


def test_generate_environment(cmake_handler):
    """Test basic generate build call with supplied environment

    Tests generate_build with environment but no off-nominal conditions, nor arguments. Delegates to the generate_build.
    """
    generate_build(({}, []), False, cmake_handler, environment={"MY_ENV_VAR": "VALUE1"})


def test_generate_exists(cmake_handler):
    """Test basic generate build call with supplied environment

    Tests generate_build with pre-existing directory.
    """
    generate_build(({}, []), True, cmake_handler)



"""
Tests for fprime.util.code_formatter
"""

from pathlib import Path

import shutil
from unittest.mock import MagicMock

import pytest
from fprime.util.code_formatter import ClangFormatter


def test_init():
    """Test initialization"""
    options = {
        "backup": True,
        "verbose": True,
        "quiet": False,
        "check": False,
        "validate_extensions": True,
    }
    formatter = ClangFormatter("clang-format", Path(".clang-format"), options)
    assert formatter is not None


def test_stage_file(tmp_path):
    """Test staging files for formatting"""
    # Create dummy files
    cpp_file = tmp_path / "test.cpp"
    cpp_file.touch()
    txt_file = tmp_path / "test.txt"
    txt_file.touch()
    non_existent_file = tmp_path / "does_not_exist.cpp"

    options = {"validate_extensions": True, "verbose": True}
    formatter = ClangFormatter("clang-format", Path(".clang-format"), options)

    # Stage files
    formatter.stage_file(cpp_file)
    formatter.stage_file(txt_file)
    formatter.stage_file(non_existent_file)

    # Check that only the valid C++ file was staged
    assert len(formatter._files_to_format) == 1
    assert formatter._files_to_format[0] == cpp_file


def test_stage_file_no_validation(tmp_path):
    """Test staging files with extension validation disabled"""
    # Create dummy files
    cpp_file = tmp_path / "test.cpp"
    cpp_file.touch()
    txt_file = tmp_path / "test.txt"
    txt_file.touch()

    options = {"validate_extensions": False}
    formatter = ClangFormatter("clang-format", Path(".clang-format"), options)

    # Stage files
    formatter.stage_file(cpp_file)
    formatter.stage_file(txt_file)

    # Check that both files were staged
    assert len(formatter._files_to_format) == 2
    assert cpp_file in formatter._files_to_format
    assert txt_file in formatter._files_to_format


def test_allow_extension(tmp_path):
    """Test allowing a new extension"""
    txt_file = tmp_path / "test.txt"
    txt_file.touch()

    options = {"validate_extensions": True}
    formatter = ClangFormatter("clang-format", Path(".clang-format"), options)
    formatter.allow_extension(".txt")
    formatter.stage_file(txt_file)

    assert len(formatter._files_to_format) == 1
    assert formatter._files_to_format[0] == txt_file


DATA_DIR = Path(__file__).parent / "data"


@pytest.fixture
def mock_build():
    """Pytest fixture for a mock build object"""
    build = MagicMock()
    build.settings = {}
    return build


@pytest.fixture
def style_file(tmp_path):
    """Pytest fixture for a style file"""
    clang_format_file = tmp_path / ".clang-format"
    clang_format_file.write_text("BasedOnStyle: LLVM\n")
    return clang_format_file


def test_execute_format(tmp_path, mock_build, style_file):
    """Test that execute formats a file"""
    malformed_src = DATA_DIR / "malformed.cpp"
    malformed_dst = tmp_path / "malformed.cpp"
    shutil.copy(malformed_src, malformed_dst)

    options = {"backup": False, "verbose": False, "quiet": True, "check": False}
    formatter = ClangFormatter("clang-format", style_file, options)
    formatter.stage_file(malformed_dst)

    result = formatter.execute(mock_build, tmp_path, ({}, []))
    assert result == 0

    well_formed_content = (DATA_DIR / "well-formed.cpp").read_text()
    assert malformed_dst.read_text() == well_formed_content


def test_execute_check_fail(tmp_path, mock_build, style_file):
    """Test that execute with --check fails on a malformed file"""
    malformed_src = DATA_DIR / "malformed.cpp"
    malformed_dst = tmp_path / "malformed.cpp"
    shutil.copy(malformed_src, malformed_dst)

    options = {"backup": False, "verbose": False, "quiet": True, "check": True}
    formatter = ClangFormatter("clang-format", style_file, options)
    formatter.stage_file(malformed_dst)

    result = formatter.execute(mock_build, tmp_path, ({}, []))
    assert result != 0


def test_execute_check_pass(tmp_path, mock_build, style_file):
    """Test that execute with --check passes on a well-formed file"""
    well_formed_src = DATA_DIR / "well-formed.cpp"
    well_formed_dst = tmp_path / "well-formed.cpp"
    shutil.copy(well_formed_src, well_formed_dst)

    options = {"backup": False, "verbose": False, "quiet": True, "check": True}
    formatter = ClangFormatter("clang-format", style_file, options)
    formatter.stage_file(well_formed_dst)

    result = formatter.execute(mock_build, tmp_path, ({}, []))
    assert result == 0



"""fprime.util.cli: CLI handling

Defines main entrypoint for fprime-util and sets up parsers for general CLI targets.

@author mstarch
"""

import argparse
import os
import re
import sys
from pathlib import Path
from typing import Callable, Dict

from fprime.fbuild.builder import GenerateException, UnableToDetectProjectException
from fprime.fbuild.cli import add_fbuild_parsers
from fprime.fbuild.target import Target
from fprime.fpp.cli import add_fpp_parsers
from fprime.util.build_helper import load_build
from fprime.util.commands import (
    run_code_format,
    run_hash_to_file,
    run_info,
    run_new,
    run_version_check,
)
from fprime.util.help_text import HelpText
from fprime.fpp.visualize import add_fpp_viz_parsers
from fprime.fpp.impl import add_fpp_impl_parsers


def utility_entry(args):
    """Entrypoint for fprime-util, main interface to F' utility"""
    parsed, cmake_args, make_args, parser, runners = parse_args(args)

    try:
        build = (
            None
            if skip_build_loading(parsed)
            else load_build(parsed, skip_build_cache_validation(parsed))
        )

        # runners is a Dict[str, Callable] of {command_name: handler_functions} pairs
        return runners[parsed.command](
            build, parsed, cmake_args, make_args, getattr(parsed, "pass_through", [])
        )

    except GenerateException as genex:
        print(
            f"[ERROR] {genex}. Partial build cache remains. Run purge to clean-up.",
            file=sys.stderr,
        )
    except UnableToDetectProjectException:
        print(f"[ERROR] Could not detect project directory for: {parsed.path}")
    except Exception as exc:
        print(f"[ERROR] {exc}", file=sys.stderr)
    return 1


def skip_build_loading(parsed):
    """Determines if the build load step should be skipped. Commands that do not require a build object
    should manually be added here by the developer.
    """
    if parsed.command == "version-check":
        return True
    return False


def skip_build_cache_validation(parsed):
    """Determines if the build cache validation step should be skipped. Commands that do not require a
    build **cache** should manually be added here by the developer.
    """
    if parsed.command in [
        "purge",
        "info",
        "format",
    ]:
        return True
    if parsed.command == "new" and parsed.new_deployment:
        return True
    return False


def add_special_parsers(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> Dict[str, Callable]:
    """Adds in CLI parsers for other commands

    Args:
        subparsers: subparsers used to create new CLI subparsers
        common: common parent parser

    Returns:
        Dictionary associating special command name to callable used to process it
    """
    # Add a search for hash function
    hash_parser = subparsers.add_parser(
        "hash-to-file",
        description=help_text.long("hash-to-file"),
        help=help_text.short("hash-to-file"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    hash_parser.add_argument(
        "hash",
        type=lambda x: int(x, 0),
        help="Assert hash value to convert to filename",
    )

    # Add a search for hash function
    subparsers.add_parser(
        "info",
        description=help_text.long("info"),
        help=help_text.short("info"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    version_parser = subparsers.add_parser(
        "version-check",
        description=help_text.long("version-check"),
        help=help_text.short("version-check"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    version_parser.add_argument(
        "--all-submodules",
        default=False,
        action="store_true",
        help="Print all submodules for version information. Defaults to only public F´ repositories, for privacy.",
    )

    # New functionality
    new_parser = subparsers.add_parser(
        "new",
        description=help_text.long("new"),
        help=help_text.short("new"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    new_parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        help="Generated files will overwrite existing ones",
    )
    new_parser.add_argument(
        "--no-venv",
        default=False,
        action="store_true",
        help="Prevent updating the virtual environment during project creation",
    )
    new_parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="Override warning about creating new deployment or component within a deployment or component",
    )
    new_exclusive = new_parser.add_argument_group(
        "'new' targets"
    ).add_mutually_exclusive_group()
    new_exclusive.add_argument(
        "--component",
        default=False,
        action="store_true",
        dest="new_component",
        help="Generate a new component",
    )
    new_exclusive.add_argument(
        "--deployment",
        default=False,
        action="store_true",
        dest="new_deployment",
        help="Generate a new deployment",
    )
    new_exclusive.add_argument(
        "--module",
        default=False,
        action="store_true",
        dest="new_module",
        help="Generate a new module",
    )
    new_exclusive.add_argument(
        "--subtopology",
        default=False,
        action="store_true",
        dest="new_subtopology",
        help="Generate a new subtopology",
    )

    # Code formatting with clang-format
    format_parser = subparsers.add_parser(
        "format",
        help=help_text.short("format"),
        description=help_text.long("format"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        conflict_handler="resolve",
    ).add_argument_group("format utility arguments")
    format_parser.add_argument(
        "--backup",
        action="store_true",
        help="Creates a .bak backup file for each formatted file",
    )
    format_parser.add_argument(
        "-q", "--quiet", action="store_true", help="Disable clang-format verbose mode"
    )
    format_parser.add_argument(
        "--force",
        action="store_true",
        help="Force all listed files to be passed to clang-format (no file extension check)",
    )
    format_parser.add_argument(
        "--allow-extension",
        action="append",
        default=[],
        help="Add a file extension to the allowed set",
    )
    format_parser.add_argument(
        "-f",
        "--files",
        nargs="*",
        default=[],
        type=Path,
        help="List of files to format",
    )
    format_parser.add_argument(
        "-d",
        "--dirs",
        nargs="*",
        default=[],
        type=Path,
        help="List of directories to format, recursively searching for files matching allowed extensions",
    )
    format_parser.add_argument(
        "--stdin", action="store_true", help="Read stdin for list of files to format"
    )
    format_parser.add_argument(
        "--check",
        action="store_true",
        help="Check if files are formatted correctly instead of formatting them",
    )
    format_parser.add_argument(
        "--pass-through",
        nargs=argparse.REMAINDER,
        default=[],
        help="If specified, --pass-through must be the last argument. Remaining arguments passed to underlying executable",
    )
    return {
        "hash-to-file": run_hash_to_file,
        "info": run_info,
        "version-check": run_version_check,
        "new": run_new,
        "format": run_code_format,
    }


def validate(parsed, unknown):
    """
    Validate rules to ensure that the args are properly consistent. This will also generate a set of validated arguments
    to pass to CMake. This allows these values to be created, defaulted, and validated in one place
    :param parsed: args to validate
    :param unknown: unknown arguments
    :return: cmake arguments to pass to CMake
    """
    # regex pattern to detect -D<CMAKE_ARGUMENT>[:<TYPE>]=<VALUE> arguments for CMake
    CMAKE_REG = re.compile(r"-D([a-zA-Z0-9_]+(?::[A-Z]+)?)=(.*)")
    cmake_args = {}
    make_args = {}
    # Check platforms for existing toolchain, unless the default is specified.
    if not hasattr(parsed, "command") or parsed.command is None:
        raise ArgValidationException("'fprime-util' not supplied sub-command argument")
    if parsed.command == "generate":
        d_args = {
            match.group(1): match.group(2)
            for match in [CMAKE_REG.match(arg) for arg in unknown]
            if match is not None
        }
        cmake_args.update(d_args)
        unknown = [arg for arg in unknown if not CMAKE_REG.match(arg)]
    # Build type only for generate, jobs only for non-generate
    elif parsed.command in [target.mnemonic for target in Target.get_all_targets()]:
        parsed.settings = None  # Force to load from cache if possible
        if parsed.jobs is not None and parsed.jobs >= 1:
            make_args["--jobs"] = parsed.jobs
    # Check if any arguments are still unknown
    if unknown:
        runnable = f"{os.path.basename(sys.argv[0])} {parsed.command}"
        msg = f"'{runnable}' supplied invalid arguments: {','.join(unknown)}"
        raise ArgValidationException(msg)
    parsed.build_cache = (
        None if parsed.build_cache is None else Path(parsed.build_cache)
    )
    return cmake_args, make_args


def parse_args(args):
    """
    Parse the arguments to the CLI. This will then enable the user to run the above listed commands via the commands.
    :param args: CLI arguments to process
    :return: parsed arguments in a Namespace
    """
    # Common parser specifying common arguments input into the utility
    common_parser = argparse.ArgumentParser(
        description="Common Parser for Common Ingredients."
    )
    common_parser.add_argument(
        "platform",
        nargs="?",
        default="default",
        help="F prime build platform (e.g. Linux, Darwin). Default specified in settings.ini",
    )
    common_parser.add_argument(
        "-r",
        "--root",
        default=None,
        help="Root of CMake project to use. May contain multiple build directories.",
    )
    common_parser.add_argument(
        "-p",
        "--path",
        default=Path.cwd(),
        type=Path,
        help="F prime directory to operate on. Default: cwd, %(default)s.",
    )
    common_parser.add_argument(
        "--build-cache",
        dest="build_cache",
        default=None,
        help="Overrides the build cache with a specific directory",
    )
    common_parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Turn on verbose output.",
    )
    common_parser.add_argument(
        "--ut", action="store_true", help="Run command against unit testing build type"
    )

    # Main parser for the whole application
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=HelpText.long("global_help"),
    )
    subparsers = parser.add_subparsers(
        description=HelpText.long("subparsers_description"), dest="command"
    )

    # Add all externally defined cli parser command to running functions
    runners = {}
    parsers = {}

    fbuild_runners, fbuild_parsers = add_fbuild_parsers(
        subparsers, common_parser, HelpText
    )
    fpp_runners, fpp_parsers = add_fpp_parsers(subparsers, common_parser)
    viz_runners, viz_parsers = add_fpp_viz_parsers(subparsers, common_parser)
    impl_runners, impl_parsers = add_fpp_impl_parsers(subparsers, common_parser)
    parsers.update(fbuild_parsers)
    parsers.update(fpp_parsers)
    parsers.update(viz_parsers)
    parsers.update(impl_parsers)
    runners.update(fbuild_runners)
    runners.update(fpp_runners)
    runners.update(viz_runners)
    runners.update(impl_runners)
    runners.update(add_special_parsers(subparsers, common_parser, HelpText))

    # Parse and prepare to run
    parsed, unknown = parser.parse_known_args(args)
    try:
        cmake_args, make_args = validate(parsed, unknown)
    except ArgValidationException as exc:
        print(f"[ERROR] {exc}", end="\n\n")
        selected_parser = parsers.get(parsed.command, parser)
        if type(selected_parser) is tuple:
            selected_parser[0].print_usage()
        else:
            selected_parser.print_usage()
        sys.exit(1)
    return parsed, cmake_args, make_args, parser, runners


class ArgValidationException(Exception):
    """An exception used for argument validation"""



"""
Tests for fprime.util.cookiecutter_wrapper
"""

import pytest

from fprime.util.cookiecutter_wrapper import find_nearest_cmake_file


@pytest.fixture
def file_structure(tmp_path):
    """Pytest fixture for a temporary file structure"""
    proj_root = tmp_path
    component_dir = proj_root / "component"
    component_subdir = component_dir / "sub"
    deployment_dir = proj_root / "deployment"

    component_dir.mkdir()
    component_subdir.mkdir()
    deployment_dir.mkdir()

    (proj_root / "project.cmake").touch()
    (proj_root / "CMakeLists.txt").touch()

    return proj_root, component_dir, component_subdir, deployment_dir


def test_find_nearest_cmake_in_component_parent(file_structure):
    """Test finding CMakeLists.txt in a component's parent directory"""
    proj_root, component_dir, component_subdir, deployment_dir = file_structure
    (component_dir / "CMakeLists.txt").touch()

    found_path = find_nearest_cmake_file(component_subdir, deployment_dir, proj_root)
    assert found_path == (component_dir / "CMakeLists.txt")


def test_find_nearest_cmake_project_cmake(file_structure):
    """Test falling back to project.cmake"""
    proj_root, component_dir, component_subdir, deployment_dir = file_structure

    found_path = find_nearest_cmake_file(component_subdir, deployment_dir, proj_root)
    assert found_path == (proj_root / "project.cmake")


def test_find_nearest_cmake_no_file(file_structure):
    """Test returning None when no file is found"""
    proj_root, component_dir, component_subdir, deployment_dir = file_structure
    (proj_root / "project.cmake").unlink()
    (proj_root / "CMakeLists.txt").unlink()

    found_path = find_nearest_cmake_file(component_subdir, deployment_dir, proj_root)
    assert found_path is None



"""
Tests the serialization and deserialization of all the types in Fw/Python/src/fprime/common/models/serialize/

Created on Jun 25, 2020
@author: hpaulson, mstarch
"""

import json
from collections.abc import Iterable

import pytest
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.time_type import TimeBase, TimeType
from fprime.common.models.serialize.type_base import BaseType, DictionaryType, ValueType
from fprime.common.models.serialize.type_exceptions import (
    AbstractMethodException,
    ArrayLengthException,
    DeserializeException,
    EnumMismatchException,
    IncorrectMembersException,
    MissingMemberException,
    NotInitializedException,
    StringSizeException,
    TypeMismatchException,
    TypeRangeException,
)

PYTHON_TESTABLE_TYPES = [
    True,
    False,
    -1,
    0,
    300,
    "abc",
    "True",
    "False",
    3.1412,
    (0, 1),
    (True, False),
    [],
    [0],
    {},
    {"abc": 123},
    {2, 4, 3},
]


def valid_values_test(type_input, valid_values, sizes):
    """Tests to be run on all types"""
    if not isinstance(sizes, Iterable):
        sizes = [sizes] * len(valid_values)

    instantiation = type_input()
    with pytest.raises(NotInitializedException):
        instantiation.serialize()
    assert instantiation.val is None
    # Setting to None is invalid
    with pytest.raises(TypeMismatchException):
        instantiation.val = None
    # Should be able to get a JSONable object that is dumpable to a JSON string
    jsonable = instantiation.to_jsonable()
    json.loads(json.dumps(jsonable))

    # Run on valid values
    for value, size in zip(valid_values, sizes):
        instantiation = type_input(val=value)
        assert instantiation.val == value
        assert instantiation.getSize() == size

        # Check assignment by value
        by_value = type_input()
        by_value.val = value
        assert by_value.val == instantiation.val, "Assignment by value has failed"
        assert by_value.getSize() == size

        # Check that the value returned by .val is also assignable to .val
        by_value.val = by_value.val

        # Check serialization and deserialization
        serialized = instantiation.serialize()
        for offset in [0, 10, 50]:
            deserializer = type_input()
            deserializer.deserialize((b" " * offset) + serialized, offset)
            assert instantiation.val == deserializer.val, "Deserialization has failed"
            assert deserializer.getSize() == size
            # Check another get/set pair and serialization of the post-deserialized object
            deserializer.val = deserializer.val
            new_serialized_bytes = deserializer.serialize()
            assert (
                serialized == new_serialized_bytes
            ), "Repeated serialization has failed"
    return instantiation


def invalid_values_test(
    type_input, invalid_values, exception_class=TypeMismatchException
):
    """Check invalid values for all types"""
    for item in invalid_values:
        # Constructor initialization
        with pytest.raises(exception_class):
            instantiation = type_input(item)
        # Value initialization
        with pytest.raises(exception_class):
            instantiation = type_input()
            instantiation.val = item
        # Deserialization problems required space
        for offset in [0, 10, 50]:
            with pytest.raises(DeserializeException):
                instantiation = type_input()
                instantiation.deserialize(b" " * offset, offset)


def ser_deser_time_test(t_base, t_context, secs, usecs):
    """
    Test serialization/deserialization of TimeType objects.

    This test function creates a time type object with the given parameters and
    then serializes it and deserializes it. Also prints it for visual inspection
    of the formatted output.

    Args:
        t_base (int): Time base for the new time type object
        t_context (int): Time context for the new time type object
        secs (int): Seconds value for the new time type object
        usecs (int): Seconds value for the new time type object
        should_err (int): True if error expected, else False

    Returns:
        True if test passed, False otherwise
    """
    val = TimeType(t_base, t_context, secs, usecs)

    buff = val.serialize()

    val2 = TimeType()
    val2.deserialize(buff, 0)

    assert val2.timeBase.value == t_base
    assert val2.timeContext == t_context
    assert val2.seconds == secs
    assert val2.useconds == usecs


def test_boolean_nominal():
    """Tests the nominal cases of a BoolType"""
    instance = valid_values_test(BoolType, [True, False], 1)
    # Make sure max size is the same as the size and can be derived from the class
    assert instance.getSize() == instance.__class__.getMaxSize()


def test_boolean_off_nominal():
    """Tests the nominal cases of a BoolType"""
    invalid_values_test(
        BoolType, filter(lambda item: not isinstance(item, bool), PYTHON_TESTABLE_TYPES)
    )


def test_int_types_nominal():
    """Tests the integer types"""
    for type_input, size in [(I8Type, 1), (I16Type, 2), (I32Type, 4), (I64Type, 8)]:
        total = pow(2, (size * 8) - 1)
        instance = valid_values_test(type_input, [0, -1, 1, -total, total - 1], size)
        # Make sure max size is the same as the size and can be derived from the class
        assert instance.getSize() == instance.__class__.getMaxSize()


def test_int_types_off_nominal():
    """Tests the integer off nominal types"""
    for type_input, size in [(I8Type, 1), (I16Type, 2), (I32Type, 4), (I64Type, 8)]:
        total = pow(2, (size * 8) - 1)
        invalid_values_test(
            type_input,
            filter(lambda item: not isinstance(item, int), PYTHON_TESTABLE_TYPES),
        )
        invalid_values_test(
            type_input, [-total - 1, total, -total * 35, total * 35], TypeRangeException
        )


def test_uint_types_nominal():
    """Tests the integer types"""
    for type_input, size in [(U8Type, 1), (U16Type, 2), (U32Type, 4), (U64Type, 8)]:
        max_int = pow(2, (size * 8)) - 1
        instance = valid_values_test(type_input, [0, 1, max_int - 1, max_int], size)
        # Make sure max size is the same as the size and can be derived from the class
        assert instance.getSize() == instance.__class__.getMaxSize()


def test_uint_types_off_nominal():
    """Tests the integer off nominal types"""
    for type_input, size in [(U8Type, 1), (U16Type, 2), (U32Type, 4), (U64Type, 8)]:
        max_int = pow(2, (size * 8)) - 1
        invalid_values_test(
            type_input,
            filter(lambda item: not isinstance(item, int), PYTHON_TESTABLE_TYPES),
        )
        invalid_values_test(
            type_input,
            [-1, -2, max_int + 1, max_int * 35, -max_int],
            TypeRangeException,
        )


def test_float_types_nominal():
    """Tests the integer types"""
    instance = valid_values_test(
        F32Type, [0.31415000557899475, 0.0, -3.141590118408203], 4
    )
    # Make sure max size is the same as the size and can be derived from the class
    assert instance.getSize() == instance.__class__.getMaxSize()
    instance = valid_values_test(
        F64Type, [0.31415000557899475, 0.0, -3.141590118408203], 8
    )
    # Make sure max size is the same as the size and can be derived from the class
    assert instance.getSize() == instance.__class__.getMaxSize()


def test_float_types_off_nominal():
    """Tests the integer off nominal types"""
    invalid_values_test(
        F32Type,
        filter(lambda item: not isinstance(item, (float, int)), PYTHON_TESTABLE_TYPES),
    )
    invalid_values_test(
        F64Type,
        filter(lambda item: not isinstance(item, (float, int)), PYTHON_TESTABLE_TYPES),
    )


def test_enum_nominal():
    """
    Tests the EnumType serialization and deserialization
    """
    members = {"MEMB1": 0, "MEMB2": 6, "MEMB3": 9}
    enum_class = EnumType.construct_type("SomeEnum", members)
    valid = ["MEMB1", "MEMB2", "MEMB3"]
    instance = valid_values_test(enum_class, valid, [4] * len(valid))
    # Make sure max size is the same as the size and can be derived from the class
    assert instance.getSize() == instance.__class__.getMaxSize()


def test_enum_off_nominal():
    """
    Tests the EnumType serialization and deserialization
    """
    members = {"MEMB1": 0, "MEMB2": 6, "MEMB3": 9}
    enum_class = EnumType.construct_type("SomeEnum", members)
    invalid = ["MEMB12", "MEMB22", "MEMB23"]
    invalid_values_test(enum_class, invalid, EnumMismatchException)
    invalid_values_test(
        enum_class,
        filter(lambda item: not isinstance(item, str), PYTHON_TESTABLE_TYPES),
    )


def test_string_nominal():
    """Tests named string types"""
    py_string = "ABC123DEF456"
    string_type = StringType.construct_type("MyFancyString", max_length=10)
    instance = valid_values_test(
        string_type, [py_string[:10], py_string[:4], py_string[:7]], [12, 6, 9]
    )
    # String type defined a max-size of 10 plus 2 for the size data
    assert instance.__class__.getMaxSize() == 10 + 2


def test_string_off_nominal():
    """Tests named string types"""
    py_string = "ABC123DEF456"
    string_type = StringType.construct_type("MyFancyString", max_length=10)
    invalid_values_test(string_type, [py_string], StringSizeException)
    invalid_values_test(
        string_type,
        filter(lambda item: not isinstance(item, str), PYTHON_TESTABLE_TYPES),
    )


def test_serializable_basic():
    """Serializable type with basic member types"""
    member_list = [
        [
            ("member1", U32Type, "%d"),
            ("member2", U32Type, "%lu"),
            ("member3", I64Type, "%lld"),
        ],
        [
            (
                "member4",
                StringType.construct_type("StringMember1", max_length=10),
                "%s",
            ),
            ("member5", StringType.construct_type("StringMember2", max_length=4), "%s"),
            ("member6", I64Type, "%lld"),
        ],
    ]
    valid_values = [
        ({"member1": 123, "member2": 456, "member3": -234}, 4 + 4 + 8, 4 + 4 + 8),
        ({"member4": "345", "member5": "abc1", "member6": 213}, 5 + 6 + 8, 12 + 6 + 8),
    ]

    for index, (members, (valid, size, max_size)) in enumerate(
        zip(member_list, valid_values)
    ):
        serializable_type = SerializableType.construct_type(
            f"BasicSerializable{index}", members
        )
        instance = valid_values_test(serializable_type, [valid], [size])
        assert (
            instance.__class__.getMaxSize() == max_size
        )  # Sum of sizes of member list


def test_serializable_basic_off_nominal():
    """Serializable type with basic member types"""
    member_list = [
        ("member1", U32Type, "%d"),
        ("member2", StringType.construct_type("StringMember1", max_length=10), "%s"),
    ]
    serializable_type = SerializableType.construct_type(
        "BasicInvalidSerializable", member_list
    )

    invalid_values = [
        ({"member5": 123, "member6": 456}, MissingMemberException),
        ({"member1": "345", "member2": 123}, TypeMismatchException),
        (
            {"member1": 345, "member2": "234", "member3": "Something"},
            IncorrectMembersException,
        ),
    ]

    for valid, exception_class in invalid_values:
        invalid_values_test(serializable_type, [valid], exception_class)
    invalid_values_test(
        serializable_type,
        filter(lambda item: not isinstance(item, dict), PYTHON_TESTABLE_TYPES),
    )


def test_serializable_advanced():
    """
    Tests the SerializableType serialization and deserialization
    """

    # First setup some classes to represent various member types ensuring that the serializable can handle them
    string_member_class = StringType.construct_type("StringMember", max_length=3)
    enum_member_class = EnumType.construct_type(
        "EnumMember", {"Option1": 0, "Option2": 6, "Option3": 9}
    )
    array_member_class = ArrayType.construct_type(
        "ArrayMember", string_member_class, 3, "%s"
    )
    sub_serializable_class = SerializableType.construct_type(
        "AdvancedSubSerializable",
        [("subfield1", U32Type), ("subfield2", array_member_class)],
    )

    field_data = [
        ("field1", string_member_class),
        ("field2", U32Type),
        ("field3", enum_member_class),
        ("field4", array_member_class),
        ("field5", sub_serializable_class),
    ]
    serializable_class = SerializableType.construct_type(
        "AdvancedSerializable", field_data
    )

    serializable1 = {
        "field1": "abc",
        "field2": 123,
        "field3": "Option2",
        "field4": ["", "123", "6"],
        "field5": {"subfield1": 3234, "subfield2": ["abc", "def", "abc"]},
    }
    instance = valid_values_test(
        serializable_class,
        [serializable1],
        [5 + 4 + 4 + (2 + 5 + 3) + (4 + (5 + 5 + 5))],
    )
    assert instance.__class__.getMaxSize() == (
        5 + 4 + 4 + (3 * 5) + (4 + (3 * 5))
    )  # Sum of sizes of member list


def test_array_type():
    """
    Tests the ArrayType serialization and deserialization
    """
    extra_ctor_args = [
        ("TestArray", I32Type, 2, "%d"),
        ("TestArray2", U8Type, 4, "%d"),
        (
            "TestArray3",
            StringType.construct_type("TestArrayString", max_length=18),
            3,
            "%s",
        ),
    ]
    values = [[32, 1], [0, 1, 2, 3], ["one", "1234", "1"]]
    sizes = [8, 4, 14]
    max_sizes = [8, 4, (2 + 18) * 3]
    for ctor_args, values, size, max_size in zip(
        extra_ctor_args, values, sizes, max_sizes
    ):
        type_input = ArrayType.construct_type(*ctor_args)
        instance = valid_values_test(type_input, [values], [size])
        assert instance.__class__.getMaxSize() == max_size


def test_array_type_off_nominal():
    """
    Test the array type for invalid values etc
    """
    type_input = ArrayType.construct_type("TestArrayPicky", I32Type, 4, "%d")
    invalid_inputs = [
        ([1, 2, 3], ArrayLengthException),
        ([1, 2, 3, 4, 5, 6], ArrayLengthException),
        (["one", "two", "three", "four"], TypeMismatchException),
    ]
    for invalid, exception_class in invalid_inputs:
        invalid_values_test(type_input, [invalid], exception_class)
    invalid_values_test(
        type_input,
        filter(lambda item: not isinstance(item, (list, tuple)), PYTHON_TESTABLE_TYPES),
    )


def test_time_type():
    """
    Tests the TimeType serialization and deserialization
    """
    TIME_SIZE = 11

    in_no_err_list = [
        (TimeBase["TB_NONE"].value, 1, 100, 999999),
        (TimeBase["TB_PROC_TIME"].value, 0xFF, 1234567, 2952),
        (TimeBase["TB_WORKSTATION_TIME"].value, 8, 1529430215, 12),
        (TimeBase["TB_SC_TIME"].value, 231, 1344230277, 123456),
        (TimeBase["TB_FPGA_TIME"].value, 78, 10395, 24556),
        (TimeBase["TB_DONT_CARE"].value, 0xB3, 12390819, 12356),
    ]

    in_err_list = [
        (10, 58, 15345, 0),
        (TimeBase["TB_NONE"].value, 1, 3, -1),
        (TimeBase["TB_WORKSTATION_TIME"].value, 1, 700000, 1234567),
    ]

    val = TimeType()
    size = val.getSize()
    assert size == TIME_SIZE
    assert val.getMaxSize() == TIME_SIZE

    for t_base, t_context, secs, usecs in in_no_err_list:
        ser_deser_time_test(t_base, t_context, secs, usecs)

    for t_base, t_context, secs, usecs in in_err_list:
        with pytest.raises(TypeRangeException):
            ser_deser_time_test(t_base, t_context, secs, usecs)


class Dummy(BaseType):
    def serialize(self):
        return "serialized"

    def deserialize(self, data, offset):
        super(Dummy, self).deserialize(data, offset)
        return "deserialized"

    def getSize(self):
        return 0

    @classmethod
    def getMaxSize(cls):
        return 902

    def to_jsonable(self):
        return {"name": "dummy"}


def test_base_type():
    with pytest.raises(TypeError) as excinfo:
        BaseType()

    assert "Can't instantiate abstract class" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo2:
        ValueType()

    assert "Can't instantiate abstract class" in str(excinfo2.value)

    d = Dummy()
    assert d.serialize() == "serialized"
    assert d.getSize() == 0
    assert Dummy.getMaxSize() == 902
    assert d.getMaxSize() == 902

    with pytest.raises(AbstractMethodException):
        # In the Dummy class above, the deserialize method
        # is set to call the super class, which is just the
        # raw abstract method, which is the only way to
        # raise an `AbstractMethodException`.
        d.deserialize("a", 0)


def test_dictionary_type_errors():
    """Ensure the dictionary type is preventing errors"""
    # Check no raw calls passing in DictionaryType
    with pytest.raises(AssertionError):
        DictionaryType.construct_type(
            DictionaryType, "MyNewString", PROPERTY="one", PROPERTY2="two"
        )

    # Check consistent field definitions: field values
    DictionaryType.construct_type(str, "MyNewString1", PROPERTY="one", PROPERTY2="two")
    with pytest.raises(AssertionError):
        DictionaryType.construct_type(
            str, "MyNewString1", PROPERTY="three", PROPERTY2="four"
        )

    # Check consistent field definitions: field names
    DictionaryType.construct_type(str, "MyNewString2", PROPERTY1="one", PROPERTY2="two")
    with pytest.raises(AssertionError):
        DictionaryType.construct_type(
            str, "MyNewString2", PROPERTY="one", PROPERTY3="two"
        )

    # Check consistent field definitions: missing field
    DictionaryType.construct_type(str, "MyNewString3", PROPERTY1="one", PROPERTY2="two")
    with pytest.raises(AssertionError):
        DictionaryType.construct_type(str, "MyNewString3", PROPERTY1="one")



"""
(test) fprime.build:

Tests the current F prime build module. Ensures that the selected singleton defines the minimum set of functions and
that they function as expected.

@author mstarch
"""

import os
import pathlib

import fprime.fbuild.builder
import fprime.fbuild.cmake
import pytest


def get_cmake_builder():
    """Gets a CMake builder for these tests

    Returns:
        CMakeBuilder for testing
    """
    return fprime.fbuild.cmake.CMakeHandler()


def get_data_dir():
    """
    Gets directory containing test-data specific to the builder being tested. This will enable new implementers, should
    there be any, to implement their own build-directory structure.

    :return:
    """
    if type(get_cmake_builder()) is fprime.fbuild.cmake.CMakeHandler:
        return os.path.join(os.path.dirname(__file__), "cmake-data")
    msg = f"Test data directory not setup for {type(get_cmake_builder())} builder class"
    raise Exception(msg)


def test_hash_finder():
    """
    Tests that the hash finder works given a known builds.
    """
    local = pathlib.Path(os.path.dirname(__file__))
    dep_dir = local / "cmake-data" / "testbuild"
    builder = fprime.fbuild.builder.Build(
        fprime.fbuild.builder.BuildType.BUILD_NORMAL, dep_dir
    )
    builder.load(
        local, build_dir=dep_dir / "build-fprime-automatic-native", skip_validation=True
    )

    assert builder.find_hashed_file(0xDEADBEEF) == ["Abc: 0xdeadbeef\n"]
    assert builder.find_hashed_file(0xC0DEC0DE) == ["HJK: 0xc0dec0de\n"]


def test_needed_functions():
    """
    Test the needed functions for the given builder. This will ensure that the public interface to the builder is
    implemented as expected.
    """
    needed_funcs = [
        "get_include_info",
        "execute_known_target",
        "get_include_locations",
        "get_fprime_configuration",
    ]
    for func in needed_funcs:
        assert hasattr(get_cmake_builder(), func)


def test_get_fprime_configuration():
    """
    Tests the given fprime configuration fetcher. Required for other portion of the system.
    """
    configs = fprime.fbuild.cmake.CMakeHandler.CMAKE_LOCATION_FIELDS
    test_data = {
        "grand-unified": (
            "/home/user11/fprime/Ref/..",
            None,
            "/home/user11/fprime/Ref/..",
        ),
        "subdir": (
            "/home/user11/Proj",
            "/home/user11/Proj/lib1;/home/user11/Proj/lib2",
            "/home/user11/Proj/fprime",
        ),
        "external": ("/home/user11/Proj", "/opt/lib1;/opt/lib2", "/opt/fprime"),
    }
    for key in test_data:
        build_dir = os.path.join(get_data_dir(), key)
        # Test all path, truth pairs
        values = get_cmake_builder().get_fprime_configuration(configs, build_dir)
        assert values == test_data[key]


def test_get_include_locations():
    """
    Test all the include locations. This will ensure that values are properly read from a cache listing. This will
    support various other portions of the system, so debug here first.
    """
    test_data = {
        "grand-unified": ["/home/user11/fprime"],
        "subdir": [
            "/home/user11/Proj",
            "/home/user11/Proj/lib1",
            "/home/user11/Proj/lib2",
            "/home/user11/Proj/fprime",
        ],
        "external": ["/home/user11/Proj", "/opt/lib1", "/opt/lib2", "/opt/fprime"],
    }
    for key in test_data:
        build_dir = os.path.join(get_data_dir(), key)
        paths = list(get_cmake_builder().get_include_locations(build_dir))
        assert paths == test_data[key]


def test_get_include_info():
    """
    Tests that the include root function gets the expected value based on the path and build-directory setups.
    """
    # Test data setup, format build_dir to tuples of path, expected result. Note: None means expect Orphan exception
    test_data = {
        "grand-unified": [
            (
                "/home/user11/fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/fprime"),
            ),
            (
                "/home/user11/fprime/Ref/SomeComp2",
                ("Ref/SomeComp2", "/home/user11/fprime"),
            ),
            (
                "/home/user11/fprime/Ref/SomeComp2/../SomeComp1",
                ("Ref/SomeComp1", "/home/user11/fprime"),
            ),
            ("/home/user11/external-sw/NachoDeploy/SomeComp3", None),
        ],
        "subdir": [
            (
                "/home/user11/Proj/fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/Proj/fprime"),
            ),
            # An insidious case where the path contains a common prefix as another possible location without being an
            # exact directory.  Notice /home/user11/Proj/fprime-something/Comps/SomeComp1 contains a non-exact common
            # prefix with the directory /home/user11/Proj/fprime. i.e. /home/user11/Proj/fprime is a prefix but
            # /home/user11/Proj/fprime/ is not.
            (
                "/home/user11/Proj/fprime-something/Comps/SomeComp1",
                ("fprime-something/Comps/SomeComp1", "/home/user11/Proj"),
            ),
            ("/home/user11/Proj/Ref/SomeComp2", ("Ref/SomeComp2", "/home/user11/Proj")),
            (
                "/home/user11/Proj/Ref/SomeComp2/../../fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/Proj/fprime"),
            ),
            ("/home/user11/external-sw/NachoDeploy/SomeComp3", None),
        ],
        "external": [
            ("/opt/fprime/Svc/SomeComp1", ("Svc/SomeComp1", "/opt/fprime")),
            ("/home/user11/Proj/Ref/SomeComp2", ("Ref/SomeComp2", "/home/user11/Proj")),
            ("/opt/something/else/external-sw/NachoDeploy/SomeComp3", None),
        ],
    }
    # Run through all the above data look for matching answers
    for key in test_data:
        build_dir = os.path.join(get_data_dir(), key)
        # Test all path, truth pairs
        for path, truth in test_data.get(key):
            if truth is None:
                with pytest.raises(fprime.fbuild.cmake.CMakeOrphanException):
                    value = get_cmake_builder().get_include_info(path, build_dir)
            else:
                value = get_cmake_builder().get_include_info(path, build_dir)
                assert value == truth


def test_find_nearest_parent_project():
    """
    This will test the ability for the system to detect valid deployment directories
    """

    test_dir = pathlib.Path(get_data_dir())
    test_data = [
        ("testbuild/subdir1/subdir2/subdir3", "testbuild/subdir1/"),
        ("testbuild/subdir1/subdir2", "testbuild/subdir1/"),
        ("testbuild/", "testbuild/"),
        ("/nonexistent/dirone/someotherpath", None),
    ]
    for path, truth in test_data:
        path = test_dir / path
        if truth is not None:
            truth = test_dir / truth
            value = fprime.fbuild.builder.Build.find_nearest_parent_project(path)
            assert value == truth
        else:
            with pytest.raises(fprime.fbuild.builder.UnableToDetectProjectException):
                fprime.fbuild.builder.Build.find_nearest_parent_project(path)



"""fprime.fbuild.target_definitions: targets definitions for fprime-util

Defines all the targets for fprime-util. Each target is a singleton that is registered into the list of all targets and
as such, each target need only be instantiated but need not be assigned to anything.

"""

from .gcovr import GcovrTarget
from .check import CheckTarget
from fprime.fbuild.check import CheckTarget
from .enumerator import (
    BasicBuildTargetEnumerator,
    MultiBuildTargetEnumerator,
    RecursiveMultiBuildTargetEnumerator,
    SpecificBuildTargetEnumerator,
    CliBuildTargetEnumerator,
)
from .target import Target, BuildSystemTarget, TargetScope
from .types import BuildType

#### "build" targets for components, deployments, unittests for both normal and testing builds ####
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build components, ports, and deployments registered in the current directory",
        scope=TargetScope.LOCAL,
        # This target will enumerate all build targets in the current directory
        build_target_enumerator=MultiBuildTargetEnumerator(),
    )
)
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build components, ports, and deployments recursively from the current directory",
        scope=TargetScope.LOCAL,
        flags={"recursive"},
        # This target will enumerate all build targets recursively from the current directory
        build_target_enumerator=RecursiveMultiBuildTargetEnumerator(),
    )
)
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build all components, ports, and deployments registered in the project",
        scope=TargetScope.GLOBAL,
        flags={"all"},
        # The build target for building all is "all"
        build_target_enumerator=SpecificBuildTargetEnumerator(["all"]),
    )
)

#### Build unittests ####
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build all components, ports, deployments, and unit tests registered in the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        flags={"ut"},
        build_type=BuildType.BUILD_TESTING,
        # This target will enumerate all build targets in the current directory
        build_target_enumerator=MultiBuildTargetEnumerator(
            fail_safe_enumerator=BasicBuildTargetEnumerator("ut_exe")
        ),
    )
)
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build components, ports, deployments, and unit tests recursively from the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        flags={"ut", "recursive"},
        build_type=BuildType.BUILD_TESTING,
        # This target will enumerate all build targets recursively from the current directory
        build_target_enumerator=RecursiveMultiBuildTargetEnumerator(),
    )
)
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build all components, ports, deployments, and unit tests registered in the project (unit test build)",
        scope=TargetScope.GLOBAL,
        flags={"all", "ut"},
        build_type=BuildType.BUILD_TESTING,
        # The build target for build all is "all"
        build_target_enumerator=SpecificBuildTargetEnumerator(["all"]),
    )
)


#### Check (build and run unittests) ####
check_multi_build_enumerator = MultiBuildTargetEnumerator(
    fail_safe_enumerator=BasicBuildTargetEnumerator("ut_exe")
)
check_multi_test_enumerator = MultiBuildTargetEnumerator(
    build_file_name=MultiBuildTargetEnumerator.TEST_BUILD_TARGETS_FILE,
    fail_safe_enumerator=BasicBuildTargetEnumerator("ut_exe"),
)

Target.register_target(
    CheckTarget(
        mnemonic="check",
        desc="Build and run unit tests in the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        build_type=BuildType.BUILD_TESTING,
        # Check uses a standard build enumerator for the build step and a test enumerator for the test step
        build_target_enumerators=[
            check_multi_build_enumerator,
            check_multi_test_enumerator,
        ],
    )
)

Target.register_target(
    CheckTarget(
        mnemonic="check",
        desc="Build and run unit tests recursively from the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        flags={"recursive"},
        build_type=BuildType.BUILD_TESTING,
        # Check uses a standard build enumerator for the build step and a test enumerator for the test step. These are
        # converted into recursive enumerators.
        build_target_enumerators=[
            RecursiveMultiBuildTargetEnumerator(
                directory_enumerator=check_multi_build_enumerator
            ),
            RecursiveMultiBuildTargetEnumerator(
                directory_enumerator=check_multi_test_enumerator
            ),
        ],
    )
)
Target.register_target(
    CheckTarget(
        mnemonic="check",
        desc="Build and run all unit tests registered in the project (unit test build)",
        scope=TargetScope.GLOBAL,
        flags={"all"},
        build_type=BuildType.BUILD_TESTING,
        # Both steps will enumerate with "all"
        build_target_enumerators=[
            SpecificBuildTargetEnumerator(["all"]),
            SpecificBuildTargetEnumerator(["all"]),
        ],
    )
)

#### Gcovr (build, run, and cover unittests) ####
Target.register_target(
    GcovrTarget(
        mnemonic="check",
        desc="Build, run, and calculate coverage for unit tests in the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        flags={"coverage"},
        build_type=BuildType.BUILD_TESTING,
        # Gcovr uses the same enumerations as check: a standard build enumerator for the build step and a test enumerator
        # for the other steps.
        build_target_enumerators=[
            check_multi_build_enumerator,
            check_multi_test_enumerator,
        ],
    )
)

Target.register_target(
    GcovrTarget(
        mnemonic="check",
        desc="Build, run, and calculate coverage for unit tests recursively from the current directory (unit test build)",
        scope=TargetScope.LOCAL,
        flags={"coverage", "recursive"},
        build_type=BuildType.BUILD_TESTING,
        # Check uses a standard build enumerator for the build step and a test enumerator for the test step. These are
        # converted into recursive enumerators.
        build_target_enumerators=[
            RecursiveMultiBuildTargetEnumerator(
                directory_enumerator=check_multi_build_enumerator
            ),
            RecursiveMultiBuildTargetEnumerator(
                directory_enumerator=check_multi_test_enumerator
            ),
        ],
    )
)

Target.register_target(
    GcovrTarget(
        mnemonic="check",
        desc="Build, run, and calculate coverage for all unit tests registered in the project (unit test build)",
        scope=TargetScope.GLOBAL,
        flags={"coverage", "all"},
        build_type=BuildType.BUILD_TESTING,
        # Both steps will enumerate with "all"
        build_target_enumerators=[
            SpecificBuildTargetEnumerator(["all"]),
            SpecificBuildTargetEnumerator(["all"]),
        ],
    )
)

#### Direct build targets ####
Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build a specific CMake target by name",
        scope=TargetScope.GLOBAL,
        flags={"target"},
        # Enumerate a specific designated target set by CLI
        build_target_enumerator=CliBuildTargetEnumerator(),
    )
)

Target.register_target(
    BuildSystemTarget(
        mnemonic="build",
        desc="Build a specific CMake target (unit test build)",
        scope=TargetScope.GLOBAL,
        flags={"ut", "target"},
        build_type=BuildType.BUILD_TESTING,
        # Enumerate a specific designated target set by CLI
        build_target_enumerator=CliBuildTargetEnumerator(),
    )
)



"""
Supplies high-level build functions to the greater fprime helper CLI. This maps from user command space to the specific
build system handler underneath.
"""

import copy
import os
import re
from pathlib import Path
from typing import Iterable, List, Union

# Forces targets into existence
import fprime.fbuild.target_definitions  # lgtm[py/unused-import]
from fprime.common.error import FprimeException
from fprime.fbuild.cmake import CMakeException, CMakeHandler
from fprime.fbuild.settings import IniSettings
from fprime.fbuild.target import Target, TargetScope
from fprime.fbuild.types import (
    AmbiguousToolchainException,
    BuildType,
    InvalidBuildCacheException,
    MissingBuildCachePath,
    NoSuchToolchainException,
    UnableToDetectProjectException,
)


class Build:
    """Represents a build configuration

    Builds in F´ consist of a build type (normal, testing), a cmake project directory, a set of settings, and a target
    platform. These are tracked as part of this Build class. This helps setup a build cache directory, load default
    settings, and track what type of build is being run.

    BuildType represents the type of build as explained in that enum type.
    Projects are an individual build of fprime, and should define the CMakeLists.txt file as a child of this
    directory. A default settings.ini file may be found here.
    Platforms represent the target hardware to build from. This is translated to the CMake toolchain file.

    After creation, a user must use invent to handle new builds (e.g. during the generation step), or load to load a
    previously generated build.

    Examples:
        To use in generation run the following code.

        build = Build(BuildType.BUILD_NORMAL, path/to/project)
        build.invent("raspberrypi")

        To use at any step after generation:

        build = Build(BuildType.BUILD_NORMAL, path/to/project)
        build.load()
    """

    VALID_CMAKE_LIST = re.compile(r"^\s*project\(.*\)", re.MULTILINE)
    CMAKE_DEFAULT_BUILD_NAME = "build-fprime-automatic-{platform}{suffix}"

    def __init__(self, build_type: BuildType, project: Path, verbose: bool = False):
        """Constructs a build object from its constituent parts

        Args:
            build_type: member of the enum BuildType specifying fprime build type
            project: path to cmake project that this build represents
        """
        self.build_type = build_type
        self.cmake_root = project
        self.settings = None
        self.platform = None
        self.build_dir = None
        self.cmake = CMakeHandler()
        self.cmake.set_verbose(verbose)

    def is_verbose(self) -> bool:
        """Returns the verbose setting of the build"""
        return self.cmake.verbose

    def invent(self, platform: str = None, build_dir: Path = None, force=False):
        """Invents a build path from a given platform

        Sets this build up as a new build that would be used as as part of a generate step. This directory must not
        already exist. If platform is None, a default will be chosen from the settings.ini file. If the settings.ini
        file does not exist, or does not specify a default_toolchain, then "native" will be used. Settings are loaded in
        this step for further uses of this build.

        build_dir is used to specify an exact build directory to use as part of this step. This allows directories to be
        specified by the caller, but is typically not used.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default

        Raises:
            InvalidBuildCacheException: a build cache already exists as it should not
        """
        self.__setup_default(platform, build_dir)
        if self.build_dir.exists():
            msg = f"{self.build_dir} already exists."
            if not force:
                raise InvalidBuildCacheException(msg)

    def load(self, platform: str = None, build_dir: Path = None, skip_validation=False):
        """Load an existing build cache

        Sets this build up from an existing build cache. This can be used after a previous run that has generated a
        build cache in order to prepare for other build steps.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default
            skip_validation: (optional) skip cache validation. Default: False, validate away!

        Raises:
            InvalidBuildCacheException: the build cache does not exist as it must
        """
        self.__setup_default(platform, build_dir)

        if skip_validation:
            return
        if (
            self.build_dir is not None
            and (self.build_dir / ".fprime-build-dir").exists()
        ):
            return

        # Message for hard-supplied --build-cache message
        if build_dir is not None:
            gen_args = f" --build-cache {build_dir}"
        else:
            gen_args = " --ut" if self.build_type == BuildType.BUILD_TESTING else ""
            gen_args += (
                " " + platform
                if platform is not None
                and platform != "native"
                and platform != "default"
                else ""
            )
        msg = f"'{self.build_dir}' is not a valid build cache. Generate this build cache with 'fprime-util generate{gen_args} ...'"
        raise InvalidBuildCacheException(
            msg,
            self.build_dir,
        )

    def get_settings(
        self,
        setting: Union[None, str, Iterable[Union[None, str]]],
        default: Union[None, str, Iterable[Union[None, str]]],
    ) -> Union[str, Iterable[str]]:
        """Fetches settings in the settings file

        Reads settings loaded from the settings file and returns them to the caller. If a single string is submitted,
        then a single string is returned. If a list of strings is submitted a list is returned. default provides default
        values to supply in the case that a setting is unavailable.

        Args:
            setting: a string or set of string settings to return
            default: a string or set of string settings to return if no setting is found

        Returns:
            a single string setting or a list of string settings to match request with defaults subbed ins
        """
        if isinstance(setting, str):
            return self.settings.get(setting, default)
        return [self.get_settings(req, back) for req, back in zip(setting, default)]

    def find_hashed_file(self, hash_value: int) -> List[str]:
        """Retrieves the file associated with a hash

        In order to reduce space and memory footprint, filenames are associated with hashes automatically as part of the
        build. This function will retrieve the file name given a has integer.

        Args:
            hash_value: hash number to lookup

        Returns:
            stored file path(s) associated with hash
        """
        hashes_file = self.build_dir / "hashes.txt"
        if not hashes_file.exists():
            msg = f"Failed to find {hashes_file}, was the build generated?"
            raise InvalidBuildCacheException(
                msg,
                self.build_dir,
            )
        with open(hashes_file) as file_handle:
            lines = filter(
                lambda line: hash_value == int(line.split(" ")[-1], 0),
                file_handle.readlines(),
            )
        return list(lines)

    def get_build_cache(self) -> Path:
        """Generates build cache path for this build

        Generates the build path for this build. This will expect a valid build path to exist unless validate is
        specified as false. A valid build cache has been created from the generate step, and thus when using this call
        as part of the generate step, validate should be set to false.

        Returns:
            Path to a build cache directory

        """
        return self.cmake_root / Build.CMAKE_DEFAULT_BUILD_NAME.format(
            platform=self.platform, suffix=self.build_type.get_suffix()
        )

    def get_build_info(self, context: Path) -> dict:
        """Constructs an informational packet about this build

        Constructs a packet that allows for users to get meta-build information. This includes: location of build, file
        and other constructs, available make targets, and other items.

        Args:
            context: contextual path to list various information about the build

        Returns:
            Build information dictionary
        """
        temp_targets = Target.get_all_targets()
        # Remove targets that are not supported given the builder and context
        temp_targets = [
            target for target in temp_targets if target.is_supported(self, context)
        ]
        # Now filter for local scope
        local_targets = [
            target
            for target in temp_targets
            if target.scope in [TargetScope.LOCAL, TargetScope.BOTH]
        ]
        global_targets = [
            target
            for target in Target.get_all_targets()
            if target.scope == TargetScope.GLOBAL
        ]
        try:
            auto_location = self.get_build_cache_path(context)
        except MissingBuildCachePath:
            auto_location = None
        return {
            "local_targets": local_targets,
            "global_targets": global_targets,
            "auto_location": auto_location,
            "build_dir": self.build_dir,
        }

    def is_project_root(self, context: Path) -> bool:
        """Check if given path represents a project root

        Args:
            context: contextual path to list various information about the build

        Returns:
            True if the context is a project, false otherwise
        """
        try:
            self.cmake.cmake_validate_source_dir(context)
            return True
        except CMakeException:
            return False

    def find_toolchain(self):
        """Locates a toolchain file in know locations

        Finds a toolchain for the given platform.  Searches in known locations for the toolchain, and compares against F
        prime provided toolchains, toolchains in libraries, and toolchains provided by project.

        Returns:
            path to CMake toolchain file or None to use builtin
        """
        assert (
            self.platform != "default"
        ), "Default toolchain should have been decided already"
        toolchain_locations = self.get_settings(
            ["framework_path", "project_root"], [None, self.cmake_root]
        )
        toolchain_locations += self.get_settings("library_locations", [])

        # If toolchain is the native target, this is supplied by CMake and we exit here.
        if self.platform == "native":
            return None
        # Otherwise, find locations of toolchain files using the specified locations from settings.
        toolchains_paths = [
            os.path.join(loc, "cmake", "toolchain", f"{self.platform}.cmake")
            for loc in toolchain_locations
            if loc is not None
        ]
        # Create a deduplicated set of toolchains
        toolchains = list(
            {
                toolchain_path
                for toolchain_path in toolchains_paths
                if os.path.exists(toolchain_path)
            }
        )
        if not toolchains:
            msg = f"Could not find toolchain file for {self.platform} at any of: {' '.join(toolchains_paths)}"
            raise NoSuchToolchainException(msg)
        if len(toolchains) > 1:
            msg = f"Found conflicting toolchain files for {self.platform} at: {' '.join(toolchains)}"
            raise AmbiguousToolchainException(msg)
        return toolchains[0]

    def get_cmake_args(self) -> dict:
        """Generates CMake arguments from project settings (settings.ini file)

        Returns:
            A dictionary of cmake settings
        """
        needed = [
            ("FPRIME_FRAMEWORK_PATH", "framework_path"),
            ("FPRIME_LIBRARY_LOCATIONS", "library_locations"),
            ("FPRIME_PROJECT_ROOT", "project_root"),
            ("FPRIME_SETTINGS_FILE", "settings_file"),
            ("FPRIME_ENVIRONMENT_FILE", "environment_file"),
            ("FPRIME_CONFIG_DIR", "config_directory"),
            ("FPRIME_INSTALL_DEST", "install_destination"),
        ]

        cmake_args = {
            cache: self.get_settings(setting, None)
            for cache, setting in needed
            if self.get_settings(setting, None) is not None
        }

        if "FPRIME_LIBRARY_LOCATIONS" in cmake_args:
            cmake_args["FPRIME_LIBRARY_LOCATIONS"] = ";".join(
                [str(location) for location in cmake_args["FPRIME_LIBRARY_LOCATIONS"]]
            )
        return cmake_args

    def get_module_name(self, path: Path):
        """Gets name of module from path"""
        return self.cmake.get_cmake_module(path, self.build_dir)

    def get_build_cache_path(self, context: Path) -> Path:
        """Get the path within the build cache associated with the given context

        Each contextual path has a matching path within the build cache that contains the outputs of the various build
        commands executed in that context. This command will return a path to that context.

        Args:
            context: contextual path to return
        """
        if not isinstance(context, Path):
            raise ValueError("Context must be a Path")
        project_relative_path = self.get_relative_path(context)
        for possible in [".", "F-Prime"]:
            possible_path = self.build_dir / possible / project_relative_path
            if possible_path.exists():
                return possible_path
        component_name = os.path.basename(context)
        msg = f"The call to add_fprime_subdirectory could not be found for {component_name}\nPlease check the CMakeLists.txt files in the parent directories of {context}"
        raise MissingBuildCachePath(msg)

    def get_relative_path(self, path: Path) -> Path:
        """Gets path relative to project"""
        relative_path, _ = self.cmake.get_include_info(path, self.build_dir)
        return Path(relative_path)

    def execute_build_target(self, build_target: str, context: Path, make_args: dict):
        """Execute a build target

        Executes a target within the build system. This will execute the target by calling into the make system. Context
        is supplied such that the system can match local targets to the global target list.

        Args:
            build_target: build system target to run as string
            context: context path for local targets
            make_args: args to supply to the build tool (make or ninja)
        """
        make_args = copy.deepcopy(make_args)
        self.cmake.execute_known_target(
            build_target,
            self.build_dir,
            context.absolute(),
            cmake_args=self.get_cmake_args(),
            make_args=make_args,
            top_target=True,
            environment=self.settings.get("environment", None),
        )

    def generate(self, user_cmake_args):
        """Generates a build given CMake arguments

        This will run a generate step of the cmake build process. This will take in any argument used/passed to CMake.

        Args:
            user_cmake_args: cmake arguments to pass into the generate step
        """
        cmake_args = {}
        try:

            def split_pair(item):
                """Process an item into a two-tuple always"""
                return tuple([*item.strip().split("=", 1), ""][:2])

            default_options_text = self.get_settings("default_cmake_options", None)
            default_options = default_options_text.split("\n")

            default_cmake_options = {
                option: value
                for (option, value) in [split_pair(item) for item in default_options]
                if option != ""
            }

            cmake_args.update(default_cmake_options)  # default_cmake_options
            cmake_args.update(user_cmake_args)  # User-supplied values from command line
            cmake_args.update(self.get_cmake_args())  # FPRIME_* values (settings.ini)

            # When the new v3 autocoder directory exists, this means we can use the new UT api and preserve the build type
            v3_autocoder_directory = Path(
                cmake_args.get("FPRIME_FRAMEWORK_PATH") / "cmake" / "autocoder"
            )
            if (
                v3_autocoder_directory.exists()
                and self.build_type == BuildType.BUILD_TESTING
            ):
                cmake_args["BUILD_TESTING"] = "ON"
                cmake_args["CMAKE_BUILD_TYPE"] = user_cmake_args.get(
                    "CMAKE_BUILD_TYPE", "Debug"
                )
            elif self.build_type == BuildType.BUILD_TESTING:
                cmake_args["CMAKE_BUILD_TYPE"] = "Testing"

            self.cmake.generate_build(
                self.cmake_root,
                self.build_dir,
                cmake_args,
                environment=self.settings.get("environment", None),
            )
        except CMakeException as cexc:
            raise GenerateException(str(cexc), cexc.exit_code) from cexc

    def purge(self):
        """Purge a build cache directory"""
        self.cmake.purge(self.build_dir)

    def purge_install(self):
        """Purge the install directory"""
        assert (
            "install_destination" in self.settings
        ), "install_destination not present in settings"
        self.cmake.purge(self.settings["install_destination"])

    def install_dest_exists(self) -> Path:
        """Check if the install destination exists and returns the path if it does"""
        assert (
            "install_destination" in self.settings
        ), "install_destination not present in settings"
        path = Path(self.settings["install_destination"])
        return path if path.exists() else None

    def refresh(self):
        """Refresh this build (i.e. the cmake build cache)

        Some build systems (e.g. CMake) require the build to be refreshed (i.e. refresh the build cache). When this
        happens it is imperative that the appropriate environment is set. For this reason, refresh is exposed as a
        helper in this layer rather than the previous pattern of calling builder.cmake.cmake_refresh_cache directly.
        """
        self.cmake.cmake_refresh_cache(
            self.build_dir, environment=self.settings.get("environment", None)
        )

    @staticmethod
    def find_nearest_parent_project(path: Path) -> Path:
        """Recurse up the directory stack looking for a valid CMake project.

        Recurse up the directory tree from the given path, looking for a project definition directory. This means it
        defines a CMakeLists.txt with a project call. This finds where the automatic build directories are allowed to
        exist.

        Notes:
            Historically, the root of the CMake project was an F' deployment. In F' > v3.2.0, this is more often an
            F' project root.

        Returns;
            Path to the nearest project directory searching up the directory tree

        Raises;
            UnableToDetectProjectException: was unable to detect a project directory
        """
        full_path = path.resolve()
        list_file = full_path / "CMakeLists.txt"
        if not full_path.parents:
            raise UnableToDetectProjectException()
        if list_file.exists():
            with open(list_file, encoding="utf8") as file_handle:
                text = file_handle.read()
            if Build.VALID_CMAKE_LIST.search(text):
                return full_path
        return Build.find_nearest_parent_project(full_path.parent)

    @staticmethod
    def get_build_list(base, build_cache=None, ignore_invalid=False):
        """Returns a list of builds that the tool will process

        Will return a list of builds the tool will process. This will be a build for each public build type unless the
        cache has been overridden.  If overridden, this will be one build pointed at that cache.

        Args:
            base: base build identified from command line. Used to get: deployment, platform,
            build_cache: (optional) path to specified build cache.
            ignore_invalid: (optional) ignore invalid build caches and add as long as they exist

        Returns:
            List of builds for public build types, or list of one for a custom build at build cache
        """
        build_types = (
            [BuildType.BUILD_CUSTOM]
            if build_cache is not None
            else BuildType.get_public_types()
        )
        builds = []
        for build_type in build_types:
            build = Build(build_type, base.cmake_root, verbose=base.cmake.verbose)
            try:
                build.load(
                    base.platform, build_dir=build_cache, skip_validation=ignore_invalid
                )
                builds.append(build)
            except InvalidBuildCacheException as error:
                # Warnings only issued when not using an explicit build cache
                if build_cache is None:
                    print(
                        f"[WARNING] Build cache '{error.cache}' invalid or not found. Skipping."
                    )
                    continue
                raise
        return builds

    def __setup_default(self, platform: str = None, build_dir: Path = None):
        """Sets up default build

        Sets this build up before determining if it is a pre-generated, or post-generated build.

        build_dir is used to specify an exact build directory to use as part of this step. This allows directories to be
        specified by the caller, but is typically not used.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default
        """
        assert self.settings is None, "Already setup it is invalid to re-setup"
        assert self.platform is None, "Already setup it is invalid to re-setup"
        assert self.build_dir is None, "Already setup it is invalid to re-setup"

        self.settings = IniSettings.load(
            self.cmake_root / "settings.ini",
            platform,
            self.build_type == BuildType.BUILD_TESTING,
        )

        if platform is not None and platform != "default":
            self.platform = platform
        elif self.build_type == BuildType.BUILD_TESTING:
            self.platform = self.settings.get("default_ut_toolchain", "native")
        else:
            self.platform = self.settings.get("default_toolchain", "native")
        self.build_dir = build_dir if build_dir is not None else self.get_build_cache()


class GenerateException(FprimeException):
    """An exception indicating generate has failed and the user may need to respond"""

    def __init__(self, message, exit_code=1):
        super().__init__(message)
        self.exit_code = exit_code



import fprime.fpp.utils.fpp_to_json.visitors.writer as Writer
import fprime.fpp.utils.fpp_to_json.node_structs as NodeStructs


def write():
    # create a new fpp module
    module = NodeStructs.Module("ExampleModule")
    module.name = "ExampleModule"

    module = Writer.ModuleWriter(module)
    print(module.open())

    # create a new fpp constant
    constant = NodeStructs.Constant("ExampleConstant")
    constant.id = "ExampleConstant"
    constant.value = "42"

    constant = Writer.ConstantWriter(constant)
    print(constant.write())

    # create a new fpp topology
    topology = NodeStructs.Topology("ExampleTopology")
    topology.name = "ExampleTopology"

    topology = Writer.TopologyWriter(topology)
    print(topology.open())

    # create a new fpp instance spec
    instanceSpec = NodeStructs.InstanceSpec("ExampleInstance")
    instanceSpec.name = "ExampleInstance"

    instanceSpec = Writer.InstanceSpecWriter(instanceSpec)
    print(instanceSpec.write())

    # close the module and topology
    print(topology.close())
    print(module.close())


if __name__ == "__main__":
    write()



"""Generic representation of autocoded array types

Created on May 29, 2020
@author: jishii
"""

from fprime.util.string_util import format_string_template

from . import serializable_type
from .type_base import DictionaryType
from .type_exceptions import (
    ArrayLengthException,
    NotInitializedException,
    TypeMismatchException,
    DeserializeException,
)


class ArrayType(DictionaryType):
    """Generic fixed-size array type representation.

    Represents a custom named type of a fixed number of like members, each of which are other types in the system.
    """

    @classmethod
    def construct_type(cls, name, member_type, length, format):
        """Constructs a sub-array type

        Constructs a new sub-type of array to represent an array of the given name, member type, length, and format
        string.

        Args:
            name: name of the array subtype
            member_type: type of the members of the array subtype
            length: length of the array subtype
            format: format string for members of the array subtype
        """
        return DictionaryType.construct_type(
            cls, name, MEMBER_TYPE=member_type, LENGTH=length, FORMAT=format
        )

    @classmethod
    def validate(cls, val):
        """Validates the values of the array"""
        if not isinstance(val, (tuple, list)):
            raise TypeMismatchException(list, type(val))
        if len(val) != cls.LENGTH:
            raise ArrayLengthException(cls.MEMBER_TYPE, cls.LENGTH, len(val))
        for i in range(cls.LENGTH):
            cls.MEMBER_TYPE.validate(val[i])

    @property
    def val(self) -> list:
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This returns such an object.

        :return dictionary of member names to python values of member keys
        """
        return None if self._val is None else [item.val for item in self._val]

    @property
    def formatted_val(self) -> list:
        """
        Format all the elements of array according to the arr_format.
        Note 1: All elements will be cast to str
        Note 2: If a member is a serializable will call serializable formatted_val
        :return a formatted array
        """
        result = []
        for item in self._val:
            if isinstance(item, (serializable_type.SerializableType, ArrayType)):
                result.append(item.formatted_val)
            else:
                result.append(format_string_template(self.FORMAT, item.val))
        return result

    @val.setter
    def val(self, val: list):
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This takes such an object and sets the
        member val list from it.

        :param val: dictionary containing python types to key names. This
        """
        self.validate(val)
        items = [self.MEMBER_TYPE(item) for item in val]
        self._val = items

    def to_jsonable(self):
        """
        JSONable array object format
        """
        return {
            "name": self.__class__.__name__,
            "type": self.__class__.__name__,
            "size": self.LENGTH,
            "format": self.FORMAT,
            "values": (
                None
                if self._val is None
                else [member.to_jsonable() for member in self._val]
            ),
        }

    def serialize(self):
        """Serialize the array by serializing the elements one by one"""
        if self.val is None:
            raise NotInitializedException(type(self))
        return b"".join([item.serialize() for item in self._val])

    def deserialize(self, data, offset):
        """Deserialize the members of the array"""
        values = []
        for field_index in range(self.LENGTH):
            try:
                item = self.MEMBER_TYPE()
                item.deserialize(data, offset)
                offset += item.getSize()
                values.append(item)
            except Exception as exc:
                raise DeserializeException(
                    f"Array index {field_index} failed to deserialize: {exc}"
                )
        self._val = values

    def getSize(self):
        """Return the size in bytes of the array"""
        return sum(item.getSize() for item in self._val)

    @classmethod
    def getMaxSize(cls):
        """Return the maximum size in bytes of the array"""
        return cls.MEMBER_TYPE.getMaxSize() * cls.LENGTH

    def __iter__(self):
        """Allow the array object to be iterated"""
        return iter(self._val)



"""fprime.common.utils: defines common utility functions to be used across sub packages

@author thomas-bc
"""

from pathlib import Path


def confirm(msg):
    """Ask user for a yes or no input after displaying the given message"""
    # Loop "forever" intended
    while True:
        confirm_input = input(msg + " (yes/no) [yes]: ")
        # Default to yes when no input
        if confirm_input.lower() in ["y", "yes", ""]:
            return True
        if confirm_input.lower() in ["n", "no"]:
            return False
        print(f"{confirm_input} is invalid.  Please use 'yes' or 'no'")


def replace_contents(filename, what, replacement, count=1):
    """Replace the first instance of what with replacement in filename"""
    changelog = Path(filename).read_text()
    with open(filename, "w") as fh:
        new_file = changelog.replace(what, replacement, count)
        fh.write(new_file)
        return new_file != changelog


def check_path_is_within_fprime_module(path: Path, is_component: bool = False):
    """Check if the current working directory is within an F prime component or deployment module

    This is done by checking if any line in the path/CMakeLists.txt file starts
    with register_fprime_* to identify both component and deployment modules or
    register_fprime_module to identify only component modules.

    This is useful to prevent users from running commands in the wrong location.
    New components can be created within a deployment module but not within a
    component module. New deployments cannot be created in deployment or component
    modules.

    Returns:
        bool: True if the path is within a valid F prime module, False otherwise
    """

    if not Path(path / "CMakeLists.txt").exists():
        return False

    with open(path / "CMakeLists.txt") as cmake_file:
        lines = cmake_file.readlines()

    # Looks for line "register_fprime_*" in CMakeLists.txt
    statement = "register_fprime_"
    if is_component:
        statement += "module"

    for line in lines:
        if line.startswith(statement):
            return True
    return False



"""fprime.util.commands: General-purpose command definitions

Defines general-purpose command processing. Those are commands that do not belong in fbuild or fpp.
Current commands include:
    - info: Prints out information about the build
    - hash-to-file: Processes hash-to-file to locate file
    - new: Creates a new component, deployment, or project
    - format: Formats code using clang-format
    - version-check: Print out versions to help debugging

@author thomas-bc
"""

import argparse
import sys
from pathlib import Path
from typing import Dict, List
import subprocess
import platform
import pkg_resources
import pip


from fprime.fbuild.builder import Build, InvalidBuildCacheException
from fprime.util.code_formatter import ClangFormatter
from .versioning import VersionException, FPRIME_PIP_PACKAGES
from fprime.util.cookiecutter_wrapper import (
    new_component,
    new_deployment,
    new_module,
    new_subtopology,
)


def run_info(
    base: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Builds and prints the informational output block

    Handles the info command using the given build as a template to discover other build types. This will load the
    information and prints it to the screen.

    Args:
        base: base build used to find other builds including information
        parsed: parsed command namespace
        _: unused cmake arguments
        __: unused make arguments
        ___: unused pass through arguments
    """
    # Roll up targets for more concise display. Dictionaries are used as their keys become a set and as of Python 3.7
    # the native dictionary maintains ordering. Thus it acts as an "ordered set" for a nice predictable display order
    # derived from the order of target definitions.
    build_infos = {}
    local_generic_targets = {}
    global_generic_targets = {}
    # Loop through available builds and harvest targets
    for build in Build.get_build_list(base, parsed.build_cache):
        build_info = build.get_build_info(Path(parsed.path))
        # Target list
        local_targets = {
            f"'{target}'": "" for target in build_info.get("local_targets", [])
        }
        global_targets = {
            f"'{target}'": "" for target in build_info.get("global_targets", [])
        }
        build_artifacts = (
            (
                build_info.get("auto_location")
                if build_info.get("auto_location") is not None
                else "N/A"
            ),
            build_info.get("build_dir", "Unknown"),
        )
        local_generic_targets.update(local_targets)
        global_generic_targets.update(global_targets)
        build_infos[build.build_type] = build_artifacts

    # Print out directory and deployment target sections
    if local_generic_targets.keys() or global_generic_targets.keys():
        print("[INFO] fprime build information:")
        print(
            f"    Available directory targets: {' '.join(local_generic_targets.keys())}"
        )
        print()
        print(
            f"    Available global targets: {' '.join(global_generic_targets.keys())}"
        )
        print(f'{"  ":-<60}')
    # Artifact locations come afterwards
    for (
        build_type,
        (build_artifact_location, global_build_cache),
    ) in build_infos.items():
        print(
            f"    {build_type.get_cmake_build_type()} build cache module directory: {build_artifact_location}"
        )
        print(
            f"    {build_type.get_cmake_build_type()} build cache: {global_build_cache}"
        )
    print()


def run_hash_to_file(
    build: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Processes hash-to-file to locate file

    Args:
        build: build to search for hashes.txt
        parsed: parsed arguments for needed for parsed.hash
        _: unused cmake arguments
        __: unused make arguments
        ___: unused pass through arguments
    """
    lines = build.find_hashed_file(parsed.hash)
    if not lines:
        msg = (
            f"Hash 0x{parsed.hash:x} not found. Do you need '--ut' for a unittest run?"
        )
        raise InvalidBuildCacheException(msg)
    print("[INFO] File(s) associated with hash 0x{:x}".format(parsed.hash))
    for line in lines:
        print("   ", line, end="")


def run_new(
    build: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Processes new command

    Args:
        build: build used to inform new_* calls
        parsed: parsed arguments
        _: unused cmake arguments
        __: unused make arguments
        ___: unused pass through arguments
    """
    if parsed.new_component:
        return new_component(build, parsed)
    if parsed.new_deployment:
        return new_deployment(build, parsed)
    if parsed.new_module:
        return new_module(build, parsed)
    if parsed.new_subtopology:
        return new_subtopology(build, parsed)
    raise NotImplementedError(
        "`fprime-util new` target is missing or not implemented. See usage (--help)."
    )


def run_code_format(
    build: Build,
    parsed: argparse.Namespace,
    _: Dict[str, str],
    __: Dict[str, str],
    ___: List[str],
):
    """Runs code formatting using clang-format

    Args:
        build: used to retrieve .clang-format file
        parsed: parsed input arguments
        __: unused cmake_args
        ___: unused make_args
        ____: unused pass-through arguments
    """
    options = {
        "quiet": parsed.quiet,
        "verbose": parsed.verbose,
        "backup": parsed.backup,
        "validate_extensions": not parsed.force,
        "check": parsed.check,
    }
    clang_formatter = ClangFormatter(
        "clang-format",
        build.settings.get("framework_path", Path(".")) / ".clang-format",
        options,
    )
    if not clang_formatter.is_supported():
        print(
            f"[ERROR] Cannot find executable: {clang_formatter.executable}. Unable to run formatting.",
            file=sys.stderr,
        )
        return 1
    # Allow requested file extensions
    for file_ext in parsed.allow_extension:
        clang_formatter.allow_extension(file_ext)
    # Stage all files that are passed through stdin, if requested
    if parsed.stdin:
        for filename in sys.stdin.read().split():
            clang_formatter.stage_file(Path(filename))
    # Stage all files that are passed through --files
    for filename in parsed.files:
        clang_formatter.stage_file(Path(filename))
    # Search for files within --dirs and stage them
    for dirname in parsed.dirs:
        dir_path = Path(dirname)
        if not dir_path.is_dir():
            print(f"[INFO] {dir_path} is not a directory. Skipping.")
            continue
        for allowed_ext in clang_formatter.allowed_extensions:
            for file in dir_path.rglob(f"*{allowed_ext}"):
                clang_formatter.stage_file(file)
    return clang_formatter.execute(build, parsed.path, ({}, parsed.pass_through))


def run_version_check(
    base: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Print out versions to help debugging

    This prints out versions of tools such as Python, CMake, and F Prime packages.
    This will also print submodules information. Only submodules in github.com are printed,
    unless all submodules are requested with --all-submodules."""

    print(f"Operating System: {platform.system()}")
    print(f"CPU Architecture: {platform.machine()}")
    print(f"Platform: {platform.platform()}")
    print(f"Python version: {platform.python_version()}")

    cmake_version = (
        subprocess.check_output(["cmake", "--version"])
        .decode("utf-8")
        .splitlines()[0]
        .split()[2]
    )
    print(f"CMake version: {cmake_version}")
    print(f"Pip version: {pip.__version__}")

    print("Pip packages:")
    for tool in FPRIME_PIP_PACKAGES:
        try:
            version = pkg_resources.get_distribution(tool).version
            print(f"    {tool}=={version}")
        except (OSError, VersionException, pkg_resources.DistributionNotFound) as exc:
            print(f"[WARNING] {exc}")

    try:
        out = (
            subprocess.check_output(
                [
                    "git",
                    "submodule",
                    "--quiet",
                    "foreach",
                    "--recursive",
                    "git remote get-url origin && git describe --tags --always",
                ]
            )
            .decode("utf-8")
            .splitlines()
        )
        fprime_version_pairs = [(out[i], out[i + 1]) for i in range(0, len(out), 2)]
        if fprime_version_pairs:
            print(f"Project submodules:")
            for remote, version in fprime_version_pairs:
                # Filter out by remotes unless all submodules are requested (for privacy reasons)
                if (
                    "github.com/nasa" in remote
                    or "github.com/fprime-community" in remote
                    or parsed.all_submodules
                ):
                    print(f"    {remote} @ {version}")
    except:
        print("[WARNING] Failed to retrieve submodule version information.")



from fprime.util.cookiecutter_wrapper import is_valid_name

# Check if the component name is valid
if is_valid_name("{{ cookiecutter.component_name }}") != "valid":
    raise ValueError(
        "Unacceptable component name. Do not use spaces or special characters"
    )
